"""
Mock devices that mimic actual data from Abode servers.

This file should be updated any time the Abode server responses
change to confirm this library can still communicate.
"""

from typing import Any, List

EMPTY_DEVICE_RESPONSE: List[Any] = []


def status_put_response_ok(devid, status):
    """Return status change response json."""
    return dict(id=devid, status=str(status))


def level_put_response_ok(devid, level):
    """Return level change response json."""
    return dict(id=devid, level=str(level))



"""Abode camera device."""

import base64
import logging
import pathlib
from shutil import copyfileobj

import requests

import jaraco

from .._itertools import single
from ..helpers import errors as ERROR
from ..helpers import timeline as TIMELINE
from ..helpers import urls
from . import base
from . import status as STATUS

log = logging.getLogger(__name__)


class Camera(base.Device):
    """Class to represent a camera device."""

    tags = (
        # motion camera
        'ir_camera',
        # motion video camera
        'ir_camcoder',
        'ipcam',
        # outdoor motion camera
        'out_view',
        # outdoor smart camera
        'vdp',
        'mini_cam',
    )
    _image_url = None
    _snapshot_base64 = None

    def capture(self):
        """Request a new camera image."""
        # Abode IP cameras use a different URL for image captures.
        if 'control_url_snapshot' in self._state:
            url = self._state['control_url_snapshot']

        elif 'control_url' in self._state:
            url = self._state['control_url']

        else:
            raise jaraco.abode.Exception(ERROR.MISSING_CONTROL_URL)

        try:
            response = self._client.send_request("put", url)

            log.debug("Capture image URL (put): %s", url)
            log.debug("Capture image response: %s", response.text)

            return True

        except jaraco.abode.Exception as exc:
            log.warning("Failed to capture image: %s", exc)

        return False

    def refresh_image(self):
        """Get the most recent camera image."""
        url = urls.TIMELINE_IMAGES_ID.format(device_id=self.id)
        response = self._client.send_request("get", url)

        log.debug("Get image URL (get): %s", url)
        log.debug("Get image response: %s", response.text)

        return self.update_image_location(response.json())

    def update_image_location(self, timeline_json):
        """Update the image location."""
        if not timeline_json:
            return False

        # If timeline_json contains a list of objects (likely), use
        # the first as it should be the "newest".
        timeline = single(timeline_json)

        # Verify that the event code is of the "CAPTURE IMAGE" event
        event_code = timeline.get('event_code')
        if event_code != TIMELINE.CAPTURE_IMAGE['event_code']:
            raise jaraco.abode.Exception(ERROR.CAM_TIMELINE_EVENT_INVALID)

        # The timeline response has an entry for "file_path" that acts as the
        # location of the image within the Abode servers.
        file_path = timeline.get('file_path')
        if not file_path:
            raise jaraco.abode.Exception(ERROR.CAM_IMAGE_REFRESH_NO_FILE)

        # Perform a "head" request for the image and look for a
        # 302 Found response
        response = self._client.send_request("head", file_path)

        if response.status_code != 302:
            log.warning(
                "Unexected response code %s with body: %s",
                str(response.status_code),
                response.text,
            )
            raise jaraco.abode.Exception(ERROR.CAM_IMAGE_UNEXPECTED_RESPONSE)

        # The response should have a location header that is the actual
        # location of the image stored on AWS
        location = response.headers.get('location')
        if not location:
            raise jaraco.abode.Exception(ERROR.CAM_IMAGE_NO_LOCATION_HEADER)

        self._image_url = location

        return True

    def stream_details_to_file(self, details, path):
        """Write the stream details to a file."""

        pathlib.Path(path).write_text(details, encoding='utf-8')
        return True

    def image_to_file(self, path, get_image=True):
        """Write the image to a file."""
        if not self.image_url or get_image:
            if not self.refresh_image():
                return False

        response = requests.get(self.image_url, stream=True)

        if response.status_code != 200:
            log.warning(
                "Unexpected response code %s when requesting image: %s",
                str(response.status_code),
                response.text,
            )
            raise jaraco.abode.Exception(ERROR.CAM_IMAGE_REQUEST_INVALID)

        with open(path, 'wb') as imgfile:
            copyfileobj(response.raw, imgfile)

        return True

    def snapshot(self):
        """Request the current camera snapshot as a base64-encoded string."""
        url = f"{urls.CAMERA_INTEGRATIONS}{self.uuid}/snapshot"

        try:
            response = self._client.send_request("post", url)
            log.debug("Camera snapshot URL (post): %s", url)
            log.debug("Camera snapshot response: %s", response.text)
        except jaraco.abode.Exception as exc:
            log.warning("Failed to get camera snapshot image: %s", exc)
            return False

        self._snapshot_base64 = response.json().get("base64Image")
        if self._snapshot_base64 is None:
            log.warning("Camera snapshot data missing")
            return False

        return True

    def snapshot_to_file(self, path, get_snapshot=True):
        """Write the snapshot image to a file."""
        if not self._snapshot_base64 or get_snapshot:
            if not self.snapshot():
                return False

        try:
            with open(path, "wb") as imgfile:
                imgfile.write(base64.b64decode(self._snapshot_base64))
        except OSError as exc:
            log.warning("Failed to write snapshot image to file: %s", exc)
            return False

        return True

    def snapshot_data_url(self, get_snapshot=True):
        """Return the snapshot image as a data url."""
        if not self._snapshot_base64 or get_snapshot:
            if not self.snapshot():
                return ""

        return f"data:image/jpeg;base64,{self._snapshot_base64}"

    def start_kvs_stream(self, path):
        """Start KVS Stream for camera."""
        url = f"{urls.CAMERA_INTEGRATIONS}{self.uuid}/kvs/stream"

        response = self._client.send_request(method="post", path=url)
        response_object = response.json()

        log.debug("Camera KVS Stream URL (post): %s", url)
        log.debug("Camera KVS Stream Response: REDACTED (due to embedded credentials)")

        if response_object['channelEndpoint'] is None:  # pragma: no cover
            raise jaraco.abode.Exception(ERROR.START_KVS_STREAM)

        log.info("Started camera %s KVS stream:", self.id)

        if status := self.stream_details_to_file(response.text, path):
            log.info(
                "Saved KVS stream endpoint data to %s for device id %s", path, self.id
            )

        return status

    def privacy_mode(self, enable):
        """Set camera privacy mode (camera on/off)."""
        if self._state['privacy']:
            privacy = '1' if enable else '0'

            path = urls.PARAMS + self.id

            camera_data = {
                'mac': self._state['camera_mac'],
                'privacy': privacy,
                'action': 'setParam',
                'id': self.id,
            }

            response = self._client.send_request(
                method="put", path=path, data=camera_data
            )
            response_object = response.json()

            log.debug("Camera Privacy Mode URL (put): %s", path)
            log.debug("Camera Privacy Mode Response: %s", response.text)

            if response_object['id'] != self.id:
                raise jaraco.abode.Exception(ERROR.SET_STATUS_DEV_ID)

            if response_object['privacy'] != str(privacy):
                raise jaraco.abode.Exception(ERROR.SET_PRIVACY_MODE)

            log.info("Set camera %s privacy mode to: %s", self.id, privacy)

            return True

        return False

    @property
    def image_url(self):
        """Get image URL."""
        return self._image_url

    @property
    def is_on(self):
        """Get camera state (assumed on)."""
        return self.status not in (STATUS.OFF, STATUS.OFFLINE)



"""Test the Abode device classes."""

import pytest

import jaraco.abode
import jaraco.abode.devices.status as STATUS
from jaraco.abode.helpers import urls

from .mock import devices as DEVICES
from .mock import login as LOGIN
from .mock import logout as LOGOUT
from .mock import oauth_claims as OAUTH_CLAIMS
from .mock import panel as PANEL
from .mock.devices import dimmer as DIMMER


class TestDimmer:
    """Test the light device with a dimmer."""

    def test_dimmer_device_properties(self, m):
        """Tests that dimmer light devices properties work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(
            urls.DEVICES,
            json=DIMMER.device(),
        )

        # Logout to reset everything
        self.client.logout()

        # Get our dimmer
        device = self.client.get_device(DIMMER.DEVICE_ID)

        # Test our device
        assert device is not None
        assert device.status == STATUS.OFF
        assert device.brightness == "0"
        assert device.has_brightness
        assert device.is_dimmable
        assert not device.has_color
        assert not device.is_color_capable
        assert not device.battery_low
        assert not device.no_response
        assert not device.is_on

        # Set up our direct device get url
        device_url = urls.DEVICE.format(id=DIMMER.DEVICE_ID)

        # Change device properties
        m.get(
            device_url,
            json=DIMMER.device(
                status=STATUS.ON,
                level=87,
                low_battery=True,
                no_response=True,
            ),
        )

        # Refesh device and test changes
        device.refresh()

        assert device.status == STATUS.ON
        assert device.brightness == "87"
        assert device.battery_low
        assert device.no_response
        assert device.is_on

    def test_dimmer_status_changes(self, m):
        """Tests that dimmer device changes work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(
            urls.DEVICES,
            json=DIMMER.device(),
        )

        # Logout to reset everything
        self.client.logout()

        # Get our dimmer
        device = self.client.get_device(DIMMER.DEVICE_ID)

        # Test that we have our device
        assert device is not None
        assert device.status == STATUS.OFF
        assert not device.is_on

        # Set up control url response
        control_url = urls.BASE + DIMMER.CONTROL_URL
        m.put(
            control_url,
            json=DEVICES.status_put_response_ok(
                devid=DIMMER.DEVICE_ID, status=int(STATUS.ON)
            ),
        )

        # Change the mode to "on"
        device.switch_on()
        assert device.status == STATUS.ON
        assert device.is_on

        # Change response
        m.put(
            control_url,
            json=DEVICES.status_put_response_ok(
                devid=DIMMER.DEVICE_ID, status=int(STATUS.OFF)
            ),
        )

        # Change the mode to "off"
        device.switch_off()
        assert device.status == STATUS.OFF
        assert not device.is_on

        # Test that an invalid status response throws exception
        m.put(
            control_url,
            json=DEVICES.status_put_response_ok(
                devid=DIMMER.DEVICE_ID, status=int(STATUS.OFF)
            ),
        )

        with pytest.raises(jaraco.abode.Exception):
            device.switch_on()



"""Test the Abode device classes."""

import pytest

import jaraco.abode
from jaraco.abode.helpers import urls

from . import mock as MOCK
from .mock import automation as AUTOMATION
from .mock import login as LOGIN
from .mock import logout as LOGOUT
from .mock import oauth_claims as OAUTH_CLAIMS
from .mock import panel as PANEL

AID_1 = '47fae27488f74f55b964a81a066c3a01'
AID_2 = '47fae27488f74f55b964a81a066c3a02'
AID_3 = '47fae27488f74f55b964a81a066c3a03'


class TestAutomation:
    """Test the automation class."""

    def test_automation_init(self, m):
        """Check the Abode automation class init's properly."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        # Set up automation
        automation_resp = AUTOMATION.get_response_ok(
            name='Auto Away',
            enabled=True,
            id=AID_1,
        )

        m.get(urls.AUTOMATION, json=automation_resp)

        # Logout to reset everything
        self.client.logout()

        # Get our specific automation
        automation = self.client.get_automation(AID_1)

        # Check automation states match
        assert automation is not None

        assert automation._state == automation_resp
        assert automation.id == str(automation_resp['id'])
        assert automation.name == automation_resp['name']
        assert automation.enabled == automation_resp['enabled']
        assert automation.desc is not None

    def test_automation_refresh(self, m):
        """Check the automation Abode class refreshes."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        # Set up automation
        resp = [
            AUTOMATION.get_response_ok(name='Test Automation', enabled=True, id=AID_1)
        ]

        m.get(urls.AUTOMATION, json=resp)

        # Set up refreshed automation
        resp_changed = [
            AUTOMATION.get_response_ok(
                name='Test Automation Changed',
                enabled=False,
                id=AID_1,
            )
        ]

        automation_id_url = urls.AUTOMATION_ID.format(id=resp[0]['id'])

        m.get(automation_id_url, json=resp_changed)

        # Logout to reset everything
        self.client.logout()

        # Get the first automation and test
        automation = self.client.get_automation(AID_1)

        # Check automation states match

        assert automation is not None
        assert automation._state == resp[0]

        # Refresh and retest
        automation.refresh()
        assert automation._state == resp_changed[0]

        # Refresh with get_automation() and test
        resp_changed = [
            AUTOMATION.get_response_ok(
                name='Test Automation Changed Again',
                enabled=True,
                id=AID_1,
            )
        ]

        m.get(automation_id_url, json=resp_changed)

        # Refresh and retest
        automation = self.client.get_automation(AID_1, refresh=True)

        assert automation._state == resp_changed[0]

        # Test refresh returning an incorrect ID throws exception
        # Set up refreshed automation
        resp_changed = [
            AUTOMATION.get_response_ok(
                name='Test Automation Changed',
                enabled=False,
                id='47fae27488f74f55b964a81a066c3a11',
            )
        ]

        m.get(automation_id_url, json=resp_changed)

        with pytest.raises(jaraco.abode.Exception):
            automation.refresh()

    def test_multiple_automations(self, m):
        """Check that multiple automations work and return correctly."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        # Set up automations
        resp = [
            AUTOMATION.get_response_ok(
                name='Test Automation One',
                enabled=True,
                id=AID_1,
            ),
            AUTOMATION.get_response_ok(
                name='Test Automation Two',
                enabled=True,
                id=AID_2,
            ),
            AUTOMATION.get_response_ok(
                name='Test Automation Three',
                enabled=True,
                id=AID_3,
            ),
        ]

        m.get(urls.AUTOMATION, json=resp)

        # Logout to reset everything
        self.client.logout()

        # Test that the automations return the correct number
        automations = self.client.get_automations()
        assert len(automations) == 3

        # Get the first automation and test

        automation_1 = self.client.get_automation(AID_1)
        assert automation_1 is not None
        assert automation_1._state == resp[0]

        automation_2 = self.client.get_automation(AID_2)
        assert automation_2 is not None
        assert automation_2._state == resp[1]

        automation_3 = self.client.get_automation(AID_3)
        assert automation_3 is not None
        assert automation_3._state == resp[2]

    def test_automation_class_reuse(self, m):
        """Check that automations reuse the same classes when refreshed."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        # Set up automations
        resp = [
            AUTOMATION.get_response_ok(
                name='Test Automation One',
                enabled=True,
                id=AID_1,
            ),
            AUTOMATION.get_response_ok(
                name='Test Automation Two',
                enabled=True,
                id=AID_2,
            ),
        ]

        m.get(urls.AUTOMATION, json=resp)

        # Logout to reset everything
        self.client.logout()

        # Test that the automations return the correct number
        automations = self.client.get_automations()
        assert len(automations) == 2

        # Get the automations and test

        automation_1 = self.client.get_automation(AID_1)
        assert automation_1 is not None
        assert automation_1._state == resp[0]

        automation_2 = self.client.get_automation(AID_2)
        assert automation_2 is not None
        assert automation_2._state == resp[1]

        # Update the automations
        resp = [
            AUTOMATION.get_response_ok(
                name='Test Automation One Changed',
                enabled=False,
                id=AID_1,
            ),
            AUTOMATION.get_response_ok(
                name='Test Automation Two Changed',
                enabled=False,
                id=AID_2,
            ),
            AUTOMATION.get_response_ok(
                name='Test Automation Three New',
                enabled=True,
                id=AID_3,
            ),
        ]

        m.get(urls.AUTOMATION, json=resp)

        # Update
        automations_changed = self.client.get_automations(refresh=True)
        assert len(automations_changed) == 3

        # Check that the original two automations have updated
        # and are using the same class
        automation_1_changed = self.client.get_automation(AID_1)
        assert automation_1_changed is not None
        assert automation_1_changed._state == resp[0]
        assert automation_1 is automation_1_changed

        automation_2_changed = self.client.get_automation(AID_2)
        assert automation_2_changed is not None
        assert automation_2_changed._state == resp[1]
        assert automation_2 is automation_2_changed

        # Check that the third new automation is correct
        automation_3 = self.client.get_automation(AID_3)
        assert automation_3 is not None
        assert automation_3._state == resp[2]

    def test_automation_enable(self, m):
        """Check that automations can change their enable state."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        # Set up automation
        resp = [
            AUTOMATION.get_response_ok(
                name='Test Automation One',
                enabled=True,
                id=AID_1,
            )
        ]

        m.get(urls.AUTOMATION, json=resp)

        # Logout to reset everything
        self.client.logout()

        # Get the automation and test

        automation = self.client.get_automation(AID_1)
        assert automation is not None
        assert automation._state == resp[0]
        assert automation.enabled

        # Set up our active state change and URL
        set_active_url = urls.AUTOMATION_ID.format(id=resp[0]['id'])

        m.patch(
            set_active_url,
            json=AUTOMATION.get_response_ok(
                name='Test Automation One',
                enabled=False,
                id=AID_1,
            ),
        )

        # Test the changed state
        automation.enable(False)
        assert not automation.enabled

        # Change the state back, this time with an array response
        m.patch(
            set_active_url,
            json=[
                AUTOMATION.get_response_ok(
                    name='Test Automation One',
                    enabled=True,
                    id=AID_1,
                )
            ],
        )

        # Test the changed state
        automation.enable(True)
        assert automation.enabled

        # Test that the response returns the wrong state
        m.patch(
            set_active_url,
            json=[
                AUTOMATION.get_response_ok(
                    name='Test Automation One',
                    enabled=True,
                    id=AID_1,
                )
            ],
        )

        with pytest.raises(jaraco.abode.Exception):
            automation.enable(False)

        # Test that the response returns the wrong id
        m.patch(
            set_active_url,
            json=[
                AUTOMATION.get_response_ok(
                    name='Test Automation One',
                    enabled=True,
                    id=AID_2,
                )
            ],
        )

        with pytest.raises(jaraco.abode.Exception):
            automation.enable(True)

    def test_automation_trigger(self, m):
        """Check that automations can be triggered."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        # Set up automation
        resp = [
            AUTOMATION.get_response_ok(
                name='Test Automation One',
                enabled=True,
                id=AID_1,
            ),
        ]

        m.get(urls.AUTOMATION, json=resp)

        # Logout to reset everything
        self.client.logout()

        # Get the automation and test

        automation = self.client.get_automation(AID_1)
        assert automation is not None

        # Set up our automation trigger reply
        set_active_url = urls.AUTOMATION_APPLY.format(id=automation.id)
        m.post(set_active_url, json=MOCK.generic_response_ok())

        # Test triggering
        automation.trigger()



"""Mock Abode Login Response."""

from . import AUTH_TOKEN, panel, user


def post_response_ok(auth_token=AUTH_TOKEN):
    """Return the successful login response json."""
    return dict(
        token=auth_token,
        expired_at='2017-06-05 00:14:12',
        initiate_screen='timeline',
        user=user.get_response_ok(),
        panel=panel.get_response_ok(),
        permissions=dict(
            premium_streaming='0',
            guest_app='0',
            family_app='0',
            multiple_accounts='1',
            google_voice='1',
            nest='1',
            alexa='1',
            ifttt='1',
            no_associates='100',
            no_contacts='2',
            no_devices='155',
            no_ipcam='100',
            no_quick_action='25',
            no_automation='75',
            media_storage='3',
            cellular_backup='0',
            cms_duration='',
            cms_included='0',
        ),
        integrations={'nest': {'is_connected': 0, 'is_home_selected': 0}},
    )


def post_response_bad_request():
    """Return the failed login response json."""
    return dict(
        code=400,
        message='Username and password do not match.',
        detail=None,
    )


def post_response_mfa_code_required():
    """Return the MFA code required login response json."""
    return dict(
        code=200,
        mfa_type='google_authenticator',
        detail=None,
    )


def post_response_bad_mfa_code():
    """Return the bad MFA code login response json."""
    return dict(code=400, message='Invalid authentication key.', detail=None)


def post_response_unknown_mfa_type():
    """Return a login response json with an unknown mfa type."""
    return dict(code=200, mfa_type='sms', detail=None)



"""Abode sensor device."""

import re

from . import base


class Sensor(base.Device):
    """Class to represent a sensor device."""

    tags = ('lm',)

    def _get_status(self, key):
        return self._state.get('statuses', {}).get(key)

    def _get_numeric_status(self, key):
        """Extract the numeric value from the statuses object."""
        value = self._get_status(key)

        if value and any(i.isdigit() for i in value):
            return float(re.sub("[^0-9.]", "", value))

    @property
    def temp(self):
        """Get device temp."""
        return self._get_numeric_status('temperature')

    @property
    def temp_unit(self):
        """Get unit of temp."""
        if '°F' in self._get_status('temperature'):
            return '°F'

        if '°C' in self._get_status('temperature'):
            return '°C'

    @property
    def humidity(self):
        """Get device humdity."""
        return self._get_numeric_status('humidity')

    @property
    def humidity_unit(self):
        """Get unit of humidity."""
        if '%' in self._get_status('humidity'):
            return '%'

    @property
    def lux(self):
        """Get device lux."""
        return self._get_numeric_status('lux')

    @property
    def lux_unit(self):
        """Get unit of lux."""
        if 'lx' in self._get_status('lux'):
            return 'lux'

    @property
    def has_temp(self):
        """Device reports temperature."""
        return self.temp is not None

    @property
    def has_humidity(self):
        """Device reports humidity level."""
        return self.humidity is not None

    @property
    def has_lux(self):
        """Device reports light lux level."""
        return self.lux is not None



"""Abode switch device."""

from typing import Tuple

from . import base
from . import status as STATUS


class Switch(base.Device):
    """Class to add switch functionality."""

    tags: Tuple[str, ...] = (
        'switch',
        'night_switch',
        'power_switch_sensor',
        'power_switch_meter',
    )

    def switch_on(self):
        """Turn the switch on."""
        self.set_status(int(STATUS.ON))
        self._state['status'] = STATUS.ON

    def switch_off(self):
        """Turn the switch off."""
        self.set_status(int(STATUS.OFF))
        self._state['status'] = STATUS.OFF

    @property
    def is_on(self):
        """
        Get switch state.

        Assume switch is on.
        """
        return self.status not in (STATUS.OFF, STATUS.OFFLINE)

    @property
    def is_dimmable(self):
        """Device dimmable."""
        return False

    @property
    def is_color_capable(self):
        """Device is color compatible."""
        # Prevents issues for switches that are specified as lights
        # in the Abode component of the Home Assistant config file
        return False

    @property
    def has_color(self):
        """Device is using color mode."""
        # Prevents issues for switches that are specified as lights
        # in the Abode component of the Home Assistant config file
        return False



"""Test the Abode device classes."""

import pytest

import jaraco.abode
from jaraco.abode.helpers import urls

from .mock import devices as DEVICES
from .mock import login as LOGIN
from .mock import logout as LOGOUT
from .mock import oauth_claims as OAUTH_CLAIMS
from .mock import panel as PANEL
from .mock.devices import alarm as ALARM


class TestAlarm:
    """Test the Alarm class."""

    def test_abode_alarm_setup(self, m):
        """Check that Abode alarm device is set up properly."""
        panel = PANEL.get_response_ok(mode='standby')
        alarm = ALARM.device(area='1', panel=panel)

        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.get(urls.DEVICES, json=DEVICES.EMPTY_DEVICE_RESPONSE)
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        alarm_device = self.client.get_alarm()

        assert alarm_device is not None

        assert alarm_device._state == alarm

    def test_alarm_device_properties(self, m):
        """Check that the abode device properties are working."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(
            urls.PANEL,
            json=PANEL.get_response_ok(
                mode='standby',
                battery=True,
                is_cellular=True,
                mac='01:AA:b3:C4:d5:66',
            ),
        )
        m.get(urls.DEVICES, json=DEVICES.EMPTY_DEVICE_RESPONSE)

        # Logout to reset everything
        self.client.logout()

        # Get alarm and test
        alarm = self.client.get_alarm()
        assert alarm.id == 'area_1'
        assert alarm is not None
        assert alarm.mode == 'standby'
        assert alarm.status == 'standby'
        assert alarm.battery
        assert alarm.is_cellular
        assert not alarm.is_on
        assert alarm.uuid == '01aab3c4d566'
        assert alarm.mac_address == '01:AA:b3:C4:d5:66'

        # Change alarm properties and state to away and test
        m.get(
            urls.PANEL,
            json=PANEL.get_response_ok(mode='away', battery=False, is_cellular=False),
        )

        # Refresh alarm and test
        alarm.refresh()

        assert alarm.id == 'area_1'
        assert alarm.mode == 'away'
        assert alarm.status == 'away'
        assert not alarm.battery
        assert not alarm.is_cellular
        assert alarm.is_on

        # Change alarm state to final on state and test
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='home'))

        # Refresh alarm and test
        alarm.refresh()
        assert alarm.mode == 'home'
        assert alarm.status == 'home'
        assert alarm.is_on

    def test_alarm_device_mode_changes(self, m):
        """Test that the abode alarm can change/report modes."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=DEVICES.EMPTY_DEVICE_RESPONSE)

        # Logout to reset everything
        self.client.logout()

        # Assert that after login we have our alarm device with standby mode
        alarm = self.client.get_alarm()

        assert alarm is not None
        assert alarm.status == 'standby'

        # Set mode URLs
        m.put(
            urls.panel_mode('1', 'standby'),
            json=PANEL.put_response_ok(mode='standby'),
        )
        m.put(
            urls.panel_mode('1', 'away'),
            json=PANEL.put_response_ok(mode='away'),
        )
        m.put(
            urls.panel_mode('1', 'home'),
            json=PANEL.put_response_ok(mode='home'),
        )

        # Set and test text based mode changes
        assert alarm.set_mode('home')
        assert alarm.mode == 'home'
        assert not alarm.is_standby
        assert alarm.is_home
        assert not alarm.is_away

        assert alarm.set_mode('away')
        assert alarm.mode == 'away'
        assert not alarm.is_standby
        assert not alarm.is_home
        assert alarm.is_away

        assert alarm.set_mode('standby')
        assert alarm.mode == 'standby'
        assert alarm.is_standby
        assert not alarm.is_home
        assert not alarm.is_away

        # Set and test direct mode changes
        assert alarm.set_home()
        assert alarm.mode == 'home'
        assert not alarm.is_standby
        assert alarm.is_home
        assert not alarm.is_away

        assert alarm.set_away()
        assert alarm.mode == 'away'
        assert not alarm.is_standby
        assert not alarm.is_home
        assert alarm.is_away

        assert alarm.set_standby()
        assert alarm.mode == 'standby'
        assert alarm.is_standby
        assert not alarm.is_home
        assert not alarm.is_away

        # Set and test default mode changes
        assert alarm.switch_off()
        assert alarm.mode == 'standby'
        assert alarm.is_standby
        assert not alarm.is_home
        assert not alarm.is_away

        self.client.set_default_mode('home')
        assert alarm.switch_on()
        assert alarm.mode == 'home'
        assert not alarm.is_standby
        assert alarm.is_home
        assert not alarm.is_away

        assert alarm.switch_off()
        assert alarm.mode == 'standby'
        assert alarm.is_standby
        assert not alarm.is_home
        assert not alarm.is_away

        self.client.set_default_mode('away')
        assert alarm.switch_on()
        assert alarm.mode == 'away'
        assert not alarm.is_standby
        assert not alarm.is_home
        assert alarm.is_away

        # Test that no mode throws exception
        with pytest.raises(jaraco.abode.Exception):
            alarm.set_mode(mode=None)

        # Test that an invalid mode throws exception
        with pytest.raises(jaraco.abode.Exception):
            alarm.set_mode('chestnuts')

        # Test that an invalid mode change response throws exception
        m.put(
            urls.panel_mode('1', 'home'),
            json=PANEL.put_response_ok(mode='away'),
        )

        with pytest.raises(jaraco.abode.Exception):
            alarm.set_mode('home')

        # Test that an invalid area in mode change response throws exception
        m.put(
            urls.panel_mode('1', 'home'),
            json=PANEL.put_response_ok(area='2', mode='home'),
        )

        with pytest.raises(jaraco.abode.Exception):
            alarm.set_mode('home')



"""Abode valve device."""

from . import status as STATUS
from .switch import Switch


class Valve(Switch):
    """Class to add valve functionality."""

    tags = ('valve',)

    def switch_on(self) -> None:
        """Open the valve."""
        self.set_status(int(STATUS.ON))
        self._state['status'] = STATUS.OPEN

    def switch_off(self) -> None:
        """Close the valve."""
        self.set_status(int(STATUS.OFF))
        self._state['status'] = STATUS.CLOSED

    @property
    def is_on(self):
        """
        Get switch state.

        Assume switch is on.
        """
        return self.status not in (STATUS.CLOSED, STATUS.OFFLINE)

    @property
    def is_dimmable(self):
        """Device dimmable."""
        return False



"""Mock Abode Door Lock Device."""

import jaraco.abode.devices.status as STATUS

DEVICE_ID = 'ZW:0000006'
CONTROL_URL = 'api/v1/control/lock/' + DEVICE_ID


def device(
    devid=DEVICE_ID,
    status=STATUS.Lock.CLOSED,
    low_battery=False,
    no_response=False,
):
    """Door lock mock device."""
    return dict(
        id=devid,
        type_tag='device_type.door_lock',
        type='Door Lock',
        name='Back Door Deadbolt',
        area='1',
        zone='7',
        sort_order='0',
        is_window='0',
        bypass='0',
        schar_24hr='0',
        sresp_mode_0='0',
        sresp_entry_0='0',
        sresp_exit_0='0',
        sresp_mode_1='0',
        sresp_entry_1='0',
        sresp_exit_1='0',
        sresp_mode_2='0',
        sresp_entry_2='0',
        sresp_exit_2='0',
        sresp_mode_3='0',
        sresp_entry_3='0',
        sresp_exit_3='0',
        version='',
        origin='abode',
        control_url=CONTROL_URL,
        deep_link=None,
        status_color='#5cb85c',
        faults={
            'low_battery': int(low_battery),
            'tempered': 0,
            'supervision': 0,
            'out_of_order': 0,
            'no_response': int(no_response),
        },
        status=status,
        statuses={'hvac_mode': None},
        status_ex='',
        actions=[
            {'label': 'Lock', 'value': 'a=1&z=7&sw=on;'},
            {'label': 'Unlock', 'value': 'a=1&z=7&sw=off;'},
        ],
        status_icons={
            'LockOpen': 'assets/icons/unlocked-red.svg',
            'LockClosed': 'assets/icons/locked-green.svg',
        },
        automation_settings=None,
        icon='assets/icons/automation-lock.svg',
    )



import logging
import warnings
from typing import Tuple

import jaraco.abode
from jaraco.classes.ancestry import iter_subclasses

from ..helpers import errors as ERROR
from ..helpers import urls
from ..state import Stateful
from . import pkg

log = logging.getLogger(__name__)


class Device(Stateful):
    """Class to represent each Abode device."""

    tags: Tuple[str, ...] = ()
    """
    Each device has a ``type_tag``, something like "device_type.door_lock".
    Each Device subclass declares the tags that it services (with the
    "device_type." prefix omitted).
    """
    _desc_t = '{name} (ID: {id}, UUID: {uuid}) - {type} - {status}'
    _url_t = urls.DEVICE

    @property
    def _control_url(self):
        if not self._state['control_url']:
            raise jaraco.abode.Exception(("Control URL required",))
        return self._state['control_url']

    def set_status(self, status) -> None:
        """Set device status."""
        response = self._client.send_request(
            method="put",
            path=self._control_url,
            data={'status': str(status)},
        )
        response_object = response.json()

        log.debug("Set Status URL (put): %s", self._control_url)
        log.debug("Set Status Response: %s", response.text)

        if response_object['id'] != self.id:
            raise jaraco.abode.Exception(ERROR.SET_STATUS_DEV_ID)

        if response_object['status'] != str(status):
            raise jaraco.abode.Exception(ERROR.SET_STATUS_STATE)

        # Note: Status returned is a string of int (e.g. "0") and not
        # the string indication normally passed for the status (e.g.
        # "LockClosed"), so can't be used to update self._state.

        log.info("Set device %s status to: %s", self.id, status)

    def set_level(self, level) -> None:
        """Set device level."""
        response = self._client.send_request(
            "put",
            self._control_url,
            data={'level': str(level)},
        )
        response_object = response.json()

        log.debug("Set Level URL (put): %s", self._control_url)
        log.debug("Set Level Response: %s", response.text)

        if response_object['id'] != self.id:
            raise jaraco.abode.Exception(ERROR.SET_STATUS_DEV_ID)

        if response_object['level'] != str(level):
            raise jaraco.abode.Exception(ERROR.SET_STATUS_STATE)

        self.update(response_object)

        log.info("Set device %s level to: %s", self.id, level)

    def get_value(self, name):
        """Get a value from the device state."""
        return self._state.get(name.lower(), {})

    @property
    def status(self):
        """Shortcut to get the generic status of a device."""
        return self.get_value('status')

    @property
    def battery_low(self):
        """Is battery level low."""
        return bool(self.get_value('faults').get('low_battery', 0))

    @property
    def no_response(self):
        """Is the device responding."""
        return bool(self.get_value('faults').get('no_response', 0))

    @property
    def out_of_order(self):
        """Is the device out of order."""
        return bool(self.get_value('faults').get('out_of_order', 0))

    @property
    def tampered(self):
        """Has the device been tampered with."""
        # 'tempered' - Typo in API?
        return bool(self.get_value('faults').get('tempered', 0))

    @property
    def name(self):
        """Get the name of this device."""
        fallback = f'{self.type} {self.id}'
        return self._state.get('name') or fallback

    @property
    def device_id(self):
        """Get the device id."""
        warnings.warn(
            "Device.device_id is deprecated. Use .id.", DeprecationWarning, stacklevel=2
        )
        return self.id

    @property
    def device_uuid(self):
        """Get the device uuid."""
        warnings.warn(
            "Device.device_uuid is deprecated. Use .uuid.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.uuid

    @classmethod
    def new(cls, state, client):
        """Create new device object for the given type."""
        pkg.import_all()
        try:
            type_tag = state['type_tag']
        except KeyError as exc:
            raise jaraco.abode.Exception(ERROR.UNABLE_TO_MAP_DEVICE) from exc

        return cls.resolve_class(type_tag)(state, client)

    @property
    def generic_type(self):
        return self.__class__.__name__.lower()

    @classmethod
    def by_type(cls):
        return {sub_cls.__name__.lower(): sub_cls for sub_cls in iter_subclasses(cls)}

    @classmethod
    def resolve_class(cls, type_tag):
        """
        >>> Device.resolve_class('device_type.povs')
        <class 'jaraco.abode.devices.binary_sensor.Motion'>
        """
        lookup = {
            f'device_type.{tag}': sub_cls
            for sub_cls in reversed(tuple(iter_subclasses(cls)))
            for tag in sub_cls.tags
        }
        return lookup.get(type_tag.lower(), Unknown)


class Unknown(Device):
    pass



"""Abode light device."""

import logging
import math

import jaraco.abode

from ..helpers import errors as ERROR
from ..helpers import urls
from .switch import Switch

log = logging.getLogger(__name__)


class ColorMode:
    on = 0
    off = 2


class Light(Switch):
    """Class for lights (dimmers)."""

    tags = ('dimmer', 'dimmer_meter', 'hue')

    def set_color_temp(self, color_temp) -> None:
        """Set device color."""
        url = urls.INTEGRATIONS + self.uuid

        color_data = {
            'action': 'setcolortemperature',
            'colorTemperature': int(color_temp),
        }

        response = self._client.send_request("post", url, data=color_data)
        response_object = response.json()

        log.debug("Set Color Temp URL (post): %s", url)
        log.debug("Set Color Temp Response: %s", response.text)

        if response_object['idForPanel'] != self.id:
            raise jaraco.abode.Exception(ERROR.SET_STATUS_DEV_ID)

        if response_object['colorTemperature'] != int(color_temp):
            log.warning(
                (
                    "Set color temp mismatch for device %s. "
                    "Request val: %s, Response val: %s "
                ),
                self.id,
                color_temp,
                response_object['colorTemperature'],
            )

            color_temp = response_object['colorTemperature']

        self.update({'statuses': {'color_temp': color_temp}})

        log.info("Set device %s color_temp to: %s", self.id, color_temp)

    def set_color(self, color) -> None:
        """Set device color."""
        url = urls.INTEGRATIONS + self.uuid

        hue, saturation = color
        color_data = {
            'action': 'setcolor',
            'hue': int(hue),
            'saturation': int(saturation),
        }

        response = self._client.send_request("post", url, data=color_data)
        response_object = response.json()

        log.debug("Set Color URL (post): %s", url)
        log.debug("Set Color Response: %s", response.text)

        if response_object['idForPanel'] != self.id:
            raise jaraco.abode.Exception(ERROR.SET_STATUS_DEV_ID)

        # Abode will sometimes return hue value off by 1 (rounding error)
        hue_comparison = math.isclose(response_object["hue"], int(hue), abs_tol=1)
        if not hue_comparison or (response_object["saturation"] != int(saturation)):
            log.warning(
                (
                    "Set color mismatch for device %s. "
                    "Request val: %s, Response val: %s "
                ),
                self.id,
                (hue, saturation),
                (response_object['hue'], response_object['saturation']),
            )

            hue = response_object['hue']
            saturation = response_object['saturation']

        self.update({'statuses': {'hue': hue, 'saturation': saturation}})

        log.info("Set device %s color to: %s", self.id, (hue, saturation))

    @property
    def brightness(self):
        """Get light brightness."""
        return self.get_value('statuses').get('level')

    @property
    def color_temp(self):
        """Get light color temp."""
        return self.get_value('statuses').get('color_temp')

    @property
    def color(self):
        """Get light color."""
        return (
            self.get_value('statuses').get('hue'),
            self.get_value('statuses').get('saturation'),
        )

    @property
    def has_brightness(self):
        """Device has brightness."""
        return self.brightness

    @property
    def has_color(self):
        """Device is using color mode."""
        return self.get_value('statuses').get('color_mode') == str(ColorMode.on)

    @property
    def is_color_capable(self):
        """Device is color compatible."""
        return 'RGB' in self.type

    @property
    def is_dimmable(self):
        """Device is dimmable."""
        return 'Dimmer' in self.type



"""Representation of an automation configured in Abode."""

import logging
import warnings
from typing import Any, Dict

import jaraco.abode

from ._itertools import single
from .helpers import errors as ERROR
from .helpers import urls
from .state import Stateful

log = logging.getLogger(__name__)


class Automation(Stateful):
    """Class for viewing and controlling automations."""

    _desc_t = '{name} (ID: {id}, Enabled: {enabled})'
    _url_t = urls.AUTOMATION_ID

    def enable(self, enable: bool):
        """Enable or disable the automation."""
        path = urls.AUTOMATION_ID.format(id=self.id)

        response = self._client.send_request(
            method="patch", path=path, data={'enabled': enable}
        )

        state: Dict[str, Any] = single(response.json())

        if state['id'] != self._state['id'] or state['enabled'] != enable:
            raise jaraco.abode.Exception(ERROR.INVALID_AUTOMATION_EDIT_RESPONSE)

        self.update(state)

        log.info("Set automation %s enable to: %s", self.name, self.enabled)
        log.debug("Automation URL (patch): %s", path)
        log.debug("Automation response: %s", response.text)

    def trigger(self):
        """Trigger the automation."""
        path = urls.AUTOMATION_APPLY.format(id=self.id)

        self._client.send_request(method="post", path=path)

        log.info("Automation triggered: %s", self.name)

    def _validate(self, state):
        if state['id'] != self.id:
            raise jaraco.abode.Exception(ERROR.INVALID_AUTOMATION_REFRESH_RESPONSE)

    @property
    def automation_id(self):
        """Get the id of the automation."""
        warnings.warn(
            "Automation.automation_id is deprecated. Use .id.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.id

    @property
    def is_enabled(self):
        """Return True if the automation is enabled."""
        warnings.warn(
            "Automation.is_enabled is deprecated. Use .enabled.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.enabled



"""Mock Abode Power Switch Sensor Device."""

import jaraco.abode.devices.status as STATUS

DEVICE_ID = 'ZW:0000000a'
CONTROL_URL = 'api/v1/control/power_switch/' + DEVICE_ID


def device(devid=DEVICE_ID, status=STATUS.OPEN, low_battery=False, no_response=False):
    """Secure barrier mock device."""
    return dict(
        id=devid,
        type_tag='device_type.secure_barrier',
        type='Secure Barrier',
        name='Garage Auto Door',
        area='1',
        zone='11',
        sort_order='0',
        is_window='0',
        bypass='0',
        schar_24hr='0',
        sresp_mode_0='0',
        sresp_entry_0='0',
        sresp_exit_0='0',
        sresp_mode_1='0',
        sresp_entry_1='0',
        sresp_exit_1='0',
        sresp_mode_2='0',
        sresp_entry_2='0',
        sresp_exit_2='0',
        sresp_mode_3='0',
        sresp_entry_3='0',
        sresp_exit_3='0',
        capture_mode=None,
        origin='abode',
        control_url=CONTROL_URL,
        deep_link=None,
        status_color='#5cb85c',
        faults={
            'low_battery': int(low_battery),
            'tempered': 0,
            'supervision': 0,
            'out_of_order': 0,
            'no_response': int(no_response),
        },
        status=status,
        statuses={'hvac_mode': None},
        status_ex='',
        actions=[
            {'label': 'Close', 'value': 'a=1&z=11&sw=off;'},
            {'label': 'Open', 'value': 'a=1&z=11&sw=on;'},
        ],
        status_icons={
            'Open': 'assets/icons/garage-door-red.svg',
            'Closed': 'assets/icons/garage-door-green.svg',
        },
        icon='assets/icons/garage-door.svg',
    )



"""Test the Abode event controller class."""

from unittest.mock import Mock, call

import pytest

import jaraco.abode
import jaraco.abode.devices.status as STATUS
import jaraco.abode.helpers.timeline as TIMELINE
from jaraco.abode.devices.binary_sensor import BinarySensor
from jaraco.abode.helpers import urls

from .mock import login as LOGIN
from .mock import logout as LOGOUT
from .mock import oauth_claims as OAUTH_CLAIMS
from .mock import panel as PANEL
from .mock.devices import door_contact as DOORCONTACT
from .mock.devices import ir_camera as IRCAMERA
from .mock.devices import secure_barrier as COVER


class TestEventController:
    """Test the event controller."""

    def test_device_id_registration(self, m):
        """Tests that we can register for device events with a device id."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=COVER.device(status=STATUS.CLOSED))

        # Logout to reset everything
        self.client.logout()

        # Get our device
        device = self.client.get_device(COVER.DEVICE_ID)
        assert device is not None

        # Get the event controller
        events = self.client.events
        assert events is not None

        # Create mock callback
        callback = Mock()

        # Register our device id
        assert events.add_device_callback(device.id, callback)

    def test_device_registration(self, m):
        """Tests that we can register for device events with a device."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=COVER.device(status=STATUS.CLOSED))

        # Logout to reset everything
        self.client.logout()

        # Get our device
        device = self.client.get_device(COVER.DEVICE_ID)
        assert device is not None

        # Get the event controller
        events = self.client.events
        assert events is not None

        # Register our device
        assert events.add_device_callback(device, lambda device: None)

    def test_device_all_unregistration(self, m):
        """Tests that we can unregister for all device events."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=COVER.device(status=STATUS.CLOSED))

        # Logout to reset everything
        self.client.logout()

        # Get our device
        device = self.client.get_device(COVER.DEVICE_ID)
        assert device is not None

        # Get the event controller
        events = self.client.events
        assert events is not None

        # Register our device
        assert events.add_device_callback(device, lambda device: None)

        # Unregister all callbacks
        assert events.remove_all_device_callbacks(device)

    def test_invalid_device(self, m):
        """Tests that invalid devices are not registered."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=COVER.device(status=STATUS.CLOSED))

        # Logout to reset everything
        self.client.logout()

        # Get our device
        device = self.client.get_device(COVER.DEVICE_ID)
        assert device is not None

        # Get the event controller
        events = self.client.events
        assert events is not None

        # Create mock callback
        callback = Mock()

        # Test that no device returns false
        assert not events.add_device_callback(None, callback)

        # Create a fake device and attempt to register that
        fake_device = BinarySensor(DOORCONTACT.device(), self.client)

        with pytest.raises(jaraco.abode.Exception):
            events.add_device_callback(fake_device, callback)

    def test_invalid_all_device_unregister(self, m):
        """Tests that invalid devices are not all unregistered."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=COVER.device(status=STATUS.CLOSED))

        # Logout to reset everything
        self.client.logout()

        # Get our device
        device = self.client.get_device(COVER.DEVICE_ID)
        assert device is not None

        # Get the event controller
        events = self.client.events
        assert events is not None

        # Test that no device returns false
        assert not events.remove_all_device_callbacks(None)

        # Create a fake device and attempt to unregister that
        fake_device = BinarySensor(DOORCONTACT.device(), self.client)

        with pytest.raises(jaraco.abode.Exception):
            events.remove_all_device_callbacks(fake_device)

    def test_event_registration(self):
        """Tests that events register correctly."""
        # Get the event controller
        events = self.client.events
        assert events is not None

        # Create mock callback
        callback = Mock()

        # Test that a valid event registers
        assert events.add_event_callback(TIMELINE.Groups.ALARM, callback)

        # Test that no event group returns false
        assert not events.add_event_callback(None, callback)

        # Test that an invalid event throws exception
        with pytest.raises(jaraco.abode.Exception):
            events.add_event_callback("lol", callback)

    def test_timeline_registration(self):
        """Tests that timeline events register correctly."""
        # Get the event controller
        events = self.client.events
        assert events is not None

        # Create mock callback
        callback = Mock()

        # Test that a valid timeline event registers
        assert events.add_timeline_callback(TIMELINE.CAPTURE_IMAGE, callback)

        # Test that no timeline event returns false
        assert not events.add_timeline_callback(None, callback)

        # Test that an invalid timeline event string throws exception
        with pytest.raises(jaraco.abode.Exception):
            events.add_timeline_callback("lol", callback)

        # Test that an invalid timeline event dict throws exception
        with pytest.raises(jaraco.abode.Exception):
            events.add_timeline_callback({"lol": "lol"}, callback)

    def test_device_callback(self, m):
        """Tests that device updates callback correctly."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=COVER.device(status=STATUS.CLOSED))

        # Logout to reset everything
        self.client.logout()

        # Get our device
        device = self.client.get_device(COVER.DEVICE_ID)
        assert device is not None

        # Get the event controller
        events = self.client.events
        assert events is not None

        callback = Mock()

        # Register our device id
        assert events.add_device_callback(device.id, callback)

        # Set up device update URL
        device_url = urls.DEVICE.format(id=COVER.DEVICE_ID)
        m.get(
            device_url,
            json=COVER.device(
                devid=COVER.DEVICE_ID,
                status=STATUS.OPEN,
                low_battery=False,
                no_response=False,
            ),
        )

        # Call our device callback method

        events._on_device_update(device.id)
        callback.assert_called_with(device)

        # Test that our device updated
        assert device.status == STATUS.OPEN

        # Test that no device ID cleanly returns
        events._on_device_update(None)

        # Test that an unknown device cleanly returns
        events._on_device_update(DOORCONTACT.DEVICE_ID)

    def test_events_callback(self):
        """Tests that event updates callback correctly."""
        # Get the event controller
        events = self.client.events
        assert events is not None

        # Create mock callbacks
        capture_callback = Mock()
        alarm_callback = Mock()

        # Register our events
        assert events.add_event_callback(TIMELINE.Groups.CAPTURE, capture_callback)

        assert events.add_event_callback(TIMELINE.Groups.ALARM, alarm_callback)

        # Call our events callback method and trigger a capture group event

        event_json = IRCAMERA.timeline_event()
        events._on_timeline_update(event_json)

        # Our capture callback should get one, but our alarm should not
        capture_callback.assert_called_with(event_json)
        alarm_callback.assert_not_called()

        # Test that an invalid event exits cleanly
        events._on_timeline_update({"invalid": "event"})

    def test_timeline_callback(self):
        """Tests that timeline updates callback correctly."""
        # Get the event controller
        events = self.client.events
        assert events is not None

        # Create mock callbacks
        all_callback = Mock()
        image_callback = Mock()
        opened_callback = Mock()

        # Register our events
        assert events.add_timeline_callback(TIMELINE.ALL, all_callback)

        assert events.add_timeline_callback(TIMELINE.CAPTURE_IMAGE, image_callback)

        assert events.add_timeline_callback(TIMELINE.OPENED, opened_callback)

        # Call our events callback method and trigger an image capture event

        event_json = IRCAMERA.timeline_event()
        events._on_timeline_update(event_json)

        # all and image callbacks should have one, opened none
        all_callback.assert_called_with(event_json)
        image_callback.assert_called_with(event_json)
        opened_callback.assert_not_called()

        # Test that an invalid event exits cleanly
        events._on_timeline_update({"invalid": "event"})

    def test_alarm_callback(self, m):
        """Tests that alarm device updates callback correctly."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=COVER.device(status=STATUS.CLOSED))

        # Logout to reset everything
        self.client.logout()

        # Get our alarm device
        alarm = self.client.get_alarm()
        assert alarm is not None

        # Get the event controller
        events = self.client.events
        assert events is not None

        callback = Mock()

        # Register our alarm for callback
        assert events.add_device_callback(alarm.id, callback)

        # Call our mode changed callback method

        events._on_mode_change('home')
        callback.assert_called_with(alarm)

        # Test that our alarm state is set properly
        assert alarm.mode == 'home'

        # Test that no mode cleanly returns
        events._on_mode_change(None)

        # Test that an unknown mode cleanly returns
        events._on_mode_change("lol")

    def test_execute_callback(self):
        """Tests that callbacks that throw exceptions don't bomb."""
        # Get the event controller
        events = self.client.events
        assert events is not None

        # Create callbacks
        def _callback(event_json):
            raise Exception("CHAOS!!!")

        # Register events callback
        assert events.add_timeline_callback(TIMELINE.CAPTURE_IMAGE, _callback)

        # Call our events callback method and trigger an image capture event

        event_json = IRCAMERA.timeline_event()
        events._on_timeline_update(event_json)

    def test_multi_device_callback(self, m):
        """Tests that multiple device updates callback correctly."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(
            urls.DEVICES,
            json=[
                COVER.device(status=STATUS.CLOSED),
                DOORCONTACT.device(status=STATUS.CLOSED),
            ],
        )

        # Logout to reset everything
        self.client.logout()

        # Get our devices
        cover = self.client.get_device(COVER.DEVICE_ID)
        assert cover is not None

        doorcontact = self.client.get_device(DOORCONTACT.DEVICE_ID)
        assert doorcontact is not None

        # Get the event controller
        events = self.client.events
        assert events is not None

        callback = Mock()

        # Register our devices
        assert events.add_device_callback([cover, doorcontact], callback)

        # Set up device update URLs
        cover_url = urls.DEVICE.format(id=COVER.DEVICE_ID)
        m.get(cover_url, json=COVER.device(status=STATUS.OPEN))

        door_url = urls.DEVICE.format(id=DOORCONTACT.DEVICE_ID)
        m.get(door_url, json=DOORCONTACT.device(status=STATUS.OPEN))

        # Call our device callback method for our cover

        events._on_device_update(cover.id)
        callback.assert_called_with(cover)

        # Test that our device updated
        assert cover.status == STATUS.OPEN

        # Test that our other device didn't update
        assert doorcontact.status == STATUS.CLOSED

        # Call our device callback method for our door contact
        events._on_device_update(doorcontact.id)
        callback.assert_has_calls([call(cover), call(doorcontact)])

        # Test that our door updated now
        assert doorcontact.status == STATUS.OPEN

    def test_multi_events_callback(self):
        """Tests that multiple event updates callback correctly."""
        # Get the event controller
        events = self.client.events
        assert events is not None

        # Create mock callback
        callback = Mock()

        # Register our events
        assert events.add_event_callback(
            [TIMELINE.Groups.ALARM, TIMELINE.Groups.CAPTURE], callback
        )

        # Call our events callback method and trigger a capture group event

        event_json = IRCAMERA.timeline_event()
        events._on_timeline_update(event_json)

        # Ensure our callback was called
        callback.assert_called_with(event_json)

    def test_multi_timeline_callback(self):
        """Tests that multiple timeline updates callback correctly."""
        # Get the event controller
        events = self.client.events
        assert events is not None

        # Create mock callback
        callback = Mock()

        # Register our events
        assert events.add_timeline_callback(
            [TIMELINE.CAPTURE_IMAGE, TIMELINE.OPENED], callback
        )

        # Call our events callback method and trigger a capture group event

        event_json = IRCAMERA.timeline_event()
        events._on_timeline_update(event_json)

        # Ensure our callback was called
        callback.assert_called_with(event_json)

    def test_automations_callback(self):
        """Tests that automation updates callback correctly."""
        # Get the event controller
        events = self.client.events
        assert events is not None

        # Create mock callbacks
        automation_callback = Mock()

        # Register our events
        assert events.add_event_callback(
            TIMELINE.Groups.AUTOMATION_EDIT, automation_callback
        )

        # Call our events callback method and trigger a capture group event

        events._on_automation_update('{}')

        # Our capture callback should get one, but our alarm should not
        automation_callback.assert_called_with('{}')



"""Mock Abode Siren Device."""

import jaraco.abode.devices.status as STATUS

DEVICE_ID = 'ZB:0005674fff'


def device(devid=DEVICE_ID, status=STATUS.ONLINE, low_battery=False, no_response=False):
    """Siren mock device."""
    return dict(
        id=devid,
        type_tag='device_type.siren',
        type='Siren',
        name='Basement Siren',
        area='1',
        zone='8',
        sort_order=None,
        is_window='',
        bypass='0',
        schar_24hr='0',
        sresp_mode_0='0',
        sresp_entry_0='0',
        sresp_exit_0='0',
        sresp_mode_1='5',
        sresp_entry_1='4',
        sresp_exit_1='0',
        sresp_mode_2='0',
        sresp_entry_2='4',
        sresp_exit_2='0',
        sresp_mode_3='0',
        sresp_entry_3='0',
        sresp_exit_3='0',
        version='852_00.00.03.05TC',
        origin='abode',
        control_url='',
        deep_link=None,
        status_color='#5cb85c',
        faults={
            'low_battery': int(low_battery),
            'tempered': 0,
            'supervision': 0,
            'out_of_order': 0,
            'no_response': int(no_response),
        },
        status=status,
        statuses={'hvac_mode': None},
        status_ex='',
        actions=[],
        status_icons=[],
        motion_event='1',
        wide_angle='0',
        icon='assets/icons/indoor-siren.svg',
    )



"""Mock Abode Panel Response."""

from typing import Any, Mapping


def get_response_ok(
    mode='standby', battery=False, is_cellular=False, mac='00:11:22:33:44:55'
) -> Mapping[str, Any]:
    """Return panel response json."""
    return dict(
        version='ABGW 0.0.2.17F ABGW-L1-XA36J 3.1.2.6.1 Z-Wave 3.95',
        report_account='5555',
        online='1',
        initialized='1',
        net_version='ABGW 0.0.2.17F',
        rf_version='ABGW-L1-XA36J',
        zigbee_version='3.1.2.6.1',
        z_wave_version='Z-Wave 3.95',
        timezone='America/New_York',
        ac_fail='0',
        battery=str(int(battery)),
        ip='192.168.1.1',
        jam='0',
        rssi='2',
        setup_zone_1='1',
        setup_zone_2='1',
        setup_zone_3='1',
        setup_zone_4='1',
        setup_zone_5='1',
        setup_zone_6='1',
        setup_zone_7='1',
        setup_zone_8='1',
        setup_zone_9='1',
        setup_zone_10='1',
        setup_gateway='1',
        setup_contacts='1',
        setup_billing='1',
        setup_users='1',
        is_cellular=str(int(is_cellular)),
        plan_set_id='1',
        dealer_id='0',
        tz_diff='-04:00',
        is_demo='0',
        rf51_version='ABGW-L1-XA36J',
        model='L1',
        mac=mac,
        xml_version='3',
        dealer_name='abode',
        id='0',
        dealer_address='2625 Middlefield Road #900 Palo Alto CA 94306',
        dealer_domain='https://my.goabode.com',
        domain_alias='https://test.goabode.com',
        dealer_support_url='https://support.goabode.com',
        app_launch_url='https://goabode.app.link/abode',
        has_wifi='0',
        mode={'area_1': mode, 'area_2': 'standby'},
    )


def put_response_ok(area='1', mode='standby'):
    """Return panel mode response json."""
    return dict(area=area, mode=mode)



"""
Mock Internal Alarm Device.

The library creates an internal device that is a standard panel response that
includes a few required device fields. This is so that we can easily translate
the panel/alarm itself into a Home Assistant device.
"""

from typing import Any, Mapping

from jaraco.abode.devices import alarm

from .. import panel as PANEL


def device(area='1', panel: Mapping[str, Any] = PANEL.get_response_ok(mode='standby')):
    """Alarm mock device."""
    return alarm.state_from_panel(panel)



"""Mock Abode Occupancy Device."""

import jaraco.abode.devices.status as STATUS

DEVICE_ID = 'RF:00000320'


def device(
    devid=DEVICE_ID,
    has_motion=False,
    low_battery=False,
    no_response=False,
    out_of_order=False,
    tampered=False,
    uuid='87a31f855e684b1ab3b87e2d13e72465',
):
    """Occupancy sensor mock device."""

    status = [STATUS.ONLINE, "Motion Detected!"][has_motion]
    statuses = dict(motion=str(int(has_motion)))
    return dict(
        id=devid,
        type_tag='device_type.povs',
        type='Occupancy',
        name='Vault motion',
        area='1',
        zone='6',
        sort_order='',
        is_window='',
        bypass='0',
        schar_24hr='0',
        sresp_24hr="5",
        sresp_mode_0='0',
        sresp_entry_0='0',
        sresp_exit_0='0',
        group_name="Motion",
        group_id="1156954",
        default_group_id="1",
        sort_id="1",
        sresp_mode_1='5',
        sresp_entry_1='4',
        sresp_exit_1='0',
        sresp_mode_2='0',
        sresp_entry_2='0',
        sresp_exit_2='0',
        sresp_mode_3='0',
        sresp_entry_3='0',
        sresp_exit_3='0',
        sresp_mode_4='0',
        sresp_entry_4='0',
        sresp_exit_4='0',
        version='0001',
        origin='abode',
        has_subscription=None,
        onboard="0",
        s2_grnt_keys="",
        s2_dsk="",
        s2_propty="",
        s2_keys_valid="",
        zwave_secure_protocol="",
        control_url='',
        deep_link=None,
        status_color='#5cb85c',
        faults={
            'low_battery': int(low_battery),
            'tempered': int(tampered),
            'supervision': 0,
            'out_of_order': int(out_of_order),
            'no_response': int(no_response),
            'jammed': 0,
            'zwave_fault': 0,
        },
        status=status,
        status_display=status,
        statuses=statuses,
        status_ex='',
        actions=[],
        status_icons=[],
        icon='assets/icons/occupancy-sensor.svg',
        sresp_trigger="0",
        sresp_restore="0",
        occupancy_timer="30",
        sensitivity="4",
        model="L1",
        is_motion_sensor=False,
        uuid=uuid,
    )



"""Test the Abode device classes."""

import pytest

import jaraco.abode
import jaraco.abode.devices.status as STATUS
from jaraco.abode.devices.light import ColorMode
from jaraco.abode.helpers import urls

from .mock import devices as DEVICES
from .mock import login as LOGIN
from .mock import logout as LOGOUT
from .mock import oauth_claims as OAUTH_CLAIMS
from .mock import panel as PANEL
from .mock.devices import hue as HUE


class TestHue:
    """Test the light device with Hue."""

    def test_hue_device_properties(self, m):
        """Tests that hue light devices properties work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(
            urls.DEVICES,
            json=HUE.device(
                color_mode=ColorMode.on,
            ),
        )

        # Logout to reset everything
        self.client.logout()

        # Get our dimmer
        device = self.client.get_device(HUE.DEVICE_ID)

        # Test our device
        assert device is not None
        assert device.status == STATUS.OFF
        assert device.brightness == "0"
        assert device.color == (60, 57)  # (hue, saturation)
        assert device.color_temp == 6536
        assert device.has_brightness
        assert device.is_dimmable
        assert device.has_color
        assert device.is_color_capable
        assert not device.battery_low
        assert not device.no_response
        assert not device.is_on

        # Set up our direct device get url
        device_url = urls.DEVICE.format(id=HUE.DEVICE_ID)

        # Change device properties
        m.get(
            device_url,
            json=HUE.device(
                status=STATUS.ON,
                level=45,
                saturation=22,
                hue=104,
                color_temp=4000,
                low_battery=True,
                no_response=True,
            ),
        )

        # Refesh device and test changes
        device.refresh()

        assert device.status == STATUS.ON
        assert device.color == (104, 22)  # (hue, saturation)
        assert device.color_temp == 4000
        assert device.has_brightness
        assert device.is_dimmable
        assert not device.has_color
        assert device.is_color_capable
        assert device.battery_low
        assert device.no_response
        assert device.is_on

    def test_hue_status_changes(self, m):
        """Tests that hue device changes work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=HUE.device(color_mode=ColorMode.on))

        # Logout to reset everything
        self.client.logout()

        # Get our hue device
        device = self.client.get_device(HUE.DEVICE_ID)

        # Test that we have our device
        assert device is not None
        assert device.status == STATUS.OFF
        assert not device.is_on

        # Set up control url response
        control_url = urls.BASE + HUE.CONTROL_URL
        m.put(
            control_url,
            json=DEVICES.status_put_response_ok(
                devid=HUE.DEVICE_ID, status=int(STATUS.ON)
            ),
        )

        # Change the mode to "on"
        device.switch_on()
        assert device.status == STATUS.ON
        assert device.is_on

        # Change response
        m.put(
            control_url,
            json=DEVICES.status_put_response_ok(
                devid=HUE.DEVICE_ID, status=int(STATUS.OFF)
            ),
        )

        # Change the mode to "off"
        device.switch_off()
        assert device.status == STATUS.OFF
        assert not device.is_on

        # Test that an invalid status response throws exception
        m.put(
            control_url,
            json=DEVICES.status_put_response_ok(
                devid=HUE.DEVICE_ID, status=int(STATUS.OFF)
            ),
        )

        with pytest.raises(jaraco.abode.Exception):
            device.switch_on()

    def test_hue_color_temp_changes(self, m):
        """Tests that hue device color temp changes work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=HUE.device(color_mode=ColorMode.on))

        # Logout to reset everything
        self.client.logout()

        # Get our hue device
        device = self.client.get_device(HUE.DEVICE_ID)

        # Test that we have our device
        assert device is not None
        assert device.status == STATUS.OFF
        assert not device.is_on
        assert device.color_temp == 6536

        # Set up integrations url response
        m.post(
            HUE.INTEGRATIONS_URL,
            json=HUE.color_temp_post_response_ok(devid=HUE.DEVICE_ID, color_temp=5554),
        )

        # Change the color temp
        device.set_color_temp(5554)
        assert device.color_temp == 5554

        # Change response
        m.post(
            HUE.INTEGRATIONS_URL,
            json=HUE.color_temp_post_response_ok(devid=HUE.DEVICE_ID, color_temp=4434),
        )

        # Change the color to something that mismatches
        device.set_color_temp(4436)

        # Assert that the color is set to the response color
        assert device.color_temp == 4434

        # Test that an invalid ID in response throws exception
        m.post(
            HUE.INTEGRATIONS_URL,
            json=HUE.color_temp_post_response_ok(
                devid=(HUE.DEVICE_ID + "23"), color_temp=4434
            ),
        )

        with pytest.raises(jaraco.abode.Exception):
            device.set_color_temp(4434)

    def test_hue_color_changes(self, m):
        """Tests that hue device color changes work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=HUE.device(color_mode=ColorMode.on))

        # Logout to reset everything
        self.client.logout()

        # Get our hue device
        device = self.client.get_device(HUE.DEVICE_ID)

        # Test that we have our device
        assert device is not None
        assert device.status == STATUS.OFF
        assert not device.is_on
        assert device.color == (60, 57)  # (hue, saturation)

        # Set up integrations url response
        m.post(
            HUE.INTEGRATIONS_URL,
            json=HUE.color_post_response_ok(devid=HUE.DEVICE_ID, hue=70, saturation=80),
        )

        # Change the color temp
        device.set_color((70, 80))
        assert device.color == (70, 80)  # (hue, saturation)

        # Change response
        m.post(
            HUE.INTEGRATIONS_URL,
            json=HUE.color_post_response_ok(devid=HUE.DEVICE_ID, hue=55, saturation=85),
        )

        # Change the color to something that mismatches
        device.set_color((44, 44))

        # Assert that the color is set to the response color
        assert device.color == (55, 85)  # (hue, saturation)

        # Test that an invalid ID in response throws exception
        m.post(
            HUE.INTEGRATIONS_URL,
            json=HUE.color_post_response_ok(
                devid=(HUE.DEVICE_ID + "23"), hue=55, saturation=85
            ),
        )

        with pytest.raises(jaraco.abode.Exception):
            device.set_color((44, 44))



BASE = 'https://my.goabode.com/'

LOGIN = '/api/auth2/login'
LOGOUT = '/api/v1/logout'

OAUTH_TOKEN = '/api/auth2/claims'

PARAMS = '/api/v1/devices_beta/'

PANEL = '/api/v1/panel'

INTEGRATIONS = '/integrations/v1/devices/'
CAMERA_INTEGRATIONS = '/integrations/v1/camera/'


def panel_mode(area, mode):
    """Create panel URL."""
    return f'/api/v1/panel/mode/{area}/{mode}'


DEVICES = '/api/v1/devices'
DEVICE = '/api/v1/devices/{id}'

AREAS = '/api/v1/areas'

SETTINGS = '/api/v1/panel/setting'
SOUNDS = '/api/v1/sounds'
SIREN = '/api/v1/siren'

AUTOMATION = '/integrations/v1/automations/'
AUTOMATION_ID = AUTOMATION + '{id}/'
AUTOMATION_APPLY = AUTOMATION_ID + 'apply'

TIMELINE_IMAGES_ID = (
    '/api/v1/timeline?device_id={device_id}&dir=next&event_label=Image+Capture&size=1'
)



"""Mock Abode Logout Response."""


def post_response_ok():
    """Return the successful logout response json."""
    return dict(code=200, message='Logout successful.')


def post_response_bad_request():
    """Return the failed logout response json."""
    return dict(code=400, message='Some logout error occurred.', detail=None)



"""Mock Abode Power Switch Meter Device."""

import jaraco.abode.devices.status as STATUS

DEVICE_ID = 'ZW:00000077'
CONTROL_URL = 'api/v1/control/power_switch/' + DEVICE_ID


def device(devid=DEVICE_ID, status=STATUS.OFF, low_battery=False, no_response=False):
    """Power switch mock device."""
    return dict(
        id=devid,
        type_tag='device_type.power_switch_meter',
        type='Power Switch Meter',
        name='Aeon Switch',
        area='1',
        zone='32',
        sort_order=None,
        is_window='',
        bypass='0',
        schar_24hr='0',
        sresp_mode_0='0',
        sresp_entry_0='0',
        sresp_exit_0='0',
        sresp_mode_1='0',
        sresp_entry_1='0',
        sresp_exit_1='0',
        sresp_mode_2='0',
        sresp_entry_2='0',
        sresp_exit_2='0',
        sresp_mode_3='0',
        sresp_entry_3='0',
        sresp_exit_3='0',
        capture_mode=None,
        origin='abode',
        control_url='api/v1/control/power_switch/ZW:00000077',
        deep_link=None,
        status_color='#5cb85c',
        faults={
            'low_battery': int(low_battery),
            'tempered': 0,
            'supervision': 0,
            'out_of_order': 0,
            'no_response': int(no_response),
        },
        status=status,
        statuses={'hvac_mode': None},
        status_ex='',
        actions=[
            {'label': 'Switch off', 'value': 'a=1&z=32&sw=off;'},
            {'label': 'Switch on', 'value': 'a=1&z=32&sw=on;'},
            {'label': 'Toggle', 'value': 'a=1&z=32&sw=toggle;'},
            {'label': 'Switch on for 5 min', 'value': 'a=1&z=32&sw=on&pd=5;'},
            {'label': 'Switch on for 10 min', 'value': 'a=1&z=32&sw=on&pd=10;'},
            {'label': 'Switch on for 15 min', 'value': 'a=1&z=32&sw=on&pd=15;'},
            {'label': 'Switch on for 20 min', 'value': 'a=1&z=32&sw=on&pd=20;'},
            {'label': 'Switch on for 25 min', 'value': 'a=1&z=32&sw=on&pd=25;'},
            {'label': 'Switch on for 30 min', 'value': 'a=1&z=32&sw=on&pd=30;'},
            {'label': 'Switch on for 45 min', 'value': 'a=1&z=32&sw=on&pd=45;'},
            {'label': 'Switch on for 1 hour', 'value': 'a=1&z=32&sw=on&pd=60;'},
            {'label': 'Switch on for 2 hours', 'value': 'a=1&z=32&sw=on&pd=120;'},
            {'label': 'Switch on for 5 hours', 'value': 'a=1&z=32&sw=on&pd=300;'},
            {'label': 'Switch on for 8 hours', 'value': 'a=1&z=32&sw=on&pd=480;'},
        ],
        status_icons=[],
        icon='assets/icons/plug.svg',
    )



"""Abode alarm device."""

import copy
import logging

import jaraco.abode

from ..helpers import errors as ERROR
from ..helpers import urls
from .switch import Switch

log = logging.getLogger(__name__)


def id(area):
    return f'area_{area}'


def state_from_panel(panel_state, area='1'):
    """Adapt panel state to alarm state."""
    alarm_state = copy.deepcopy(panel_state)
    alarm_state['name'] = 'Abode Alarm'
    alarm_state['id'] = id(area)
    alarm_state['type'] = 'Alarm'
    alarm_state['type_tag'] = 'device_type.alarm'
    return alarm_state


def create_alarm(panel_json, abode, area='1'):
    """Create a new alarm device from a panel response."""
    return Alarm(state_from_panel(panel_json), abode, area)


class Alarm(Switch):
    """Class to represent the Abode alarm as a device."""

    tags = ('alarm',)
    all_modes = 'away', 'standby', 'home'

    def __init__(self, json_obj, abode, area='1'):
        """Set up Abode alarm device."""
        super().__init__(json_obj, abode)
        self._area = area

    def set_mode(self, mode):
        """Set Abode alarm mode."""
        if not mode:
            raise jaraco.abode.Exception(ERROR.MISSING_ALARM_MODE)

        if mode.lower() not in self.all_modes:
            raise jaraco.abode.Exception(ERROR.INVALID_ALARM_MODE)

        mode = mode.lower()

        response = self._client.send_request("put", urls.panel_mode(self._area, mode))

        log.debug("Set Alarm Home URL (put): %s", urls.panel_mode(self._area, mode))
        log.debug("Set Alarm Home Response: %s", response.text)

        response_object = response.json()

        if response_object['area'] != self._area:
            raise jaraco.abode.Exception(ERROR.SET_MODE_AREA)

        if response_object['mode'] != mode:
            raise jaraco.abode.Exception(ERROR.SET_MODE_MODE)

        self._state['mode'][(self.id)] = response_object['mode']

        log.info("Set alarm %s mode to: %s", self.id, response_object['mode'])

        return True

    def set_home(self):
        """Arm Abode to home mode."""
        return self.set_mode('home')

    def set_away(self):
        """Arm Abode to home mode."""
        return self.set_mode('away')

    def set_standby(self):
        """Arm Abode to stay mode."""
        return self.set_mode('standby')

    def switch_on(self):
        """Arm Abode to default mode."""
        return self.set_mode(self._client.default_mode)

    def switch_off(self):
        """Arm Abode to home mode."""
        return self.set_standby()

    def refresh(self, url=urls.PANEL):
        """Refresh the alarm device."""
        state = super().refresh(url)

        self._client._panel.update(state)

        return state

    def update(self, state):
        super().update(state_from_panel(state, area=self._area))

    @property
    def is_on(self):
        """Is alarm armed."""
        return self.mode in ('home', 'away')

    @property
    def is_standby(self):
        """Is alarm in standby mode."""
        return self.mode == 'standby'

    @property
    def is_home(self):
        """Is alarm in home mode."""
        return self.mode == 'home'

    @property
    def is_away(self):
        """Is alarm in away mode."""
        return self.mode == 'away'

    @property
    def mode(self):
        """Get alarm mode."""
        mode = self.get_value('mode').get(self.id, None)

        return mode.lower()

    @property
    def status(self):
        """To match existing property."""
        return self.mode

    @property
    def battery(self):
        """Return true if base station on battery backup."""
        return int(self._state.get('battery', '0')) == 1

    @property
    def is_cellular(self):
        """Return true if base station on cellular backup."""
        return int(self._state.get('is_cellular', '0')) == 1

    @property
    def mac_address(self):
        """Get the hub mac address."""
        return self.mac

    @property
    def uuid(self):
        return self.mac.replace(':', '').lower()



import builtins

import requests


class Exception(builtins.Exception):
    """Class to throw general abode exception."""

    def __init__(self, error):
        super().__init__(*error)

    @property
    def errcode(self):
        code, _ = self.args
        return code

    @property
    def message(self):
        _, message = self.args
        return message


class AuthenticationException(Exception):
    """Class to throw authentication exception."""

    @classmethod
    def raise_for(cls, response):
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as exc:
            raise cls((response.status_code, cls.best_message(response))) from exc

    @staticmethod
    def best_message(response):
        if response.headers.get('Content-Type') == 'application/json':
            return response.json()['message']
        return response.text


class SocketIOException(Exception):
    """Class to throw SocketIO Error exception."""

    def __init__(self, error, details):
        super().__init__(error)
        self.details = details



"""Abode binary sensor device."""

from . import base
from . import status as STATUS


class BinarySensor(base.Device):
    """Class to represent an on / off, online/offline sensor."""

    @property
    def is_on(self):
        """
        Get sensor state.

        Assume offline or open (worst case).
        """
        return self.status not in (
            STATUS.OFF,
            STATUS.OFFLINE,
            STATUS.CLOSED,
        )


class Connectivity(BinarySensor):
    tags = (
        'glass',
        'keypad',
        'remote_controller',
        'siren',
        # status display
        'bx',
        # moisture
        'water_sensor',
    )


class Motion(BinarySensor):
    tags = (
        'pir',
        'povs',
    )

    @property
    def is_on(self):
        if self.type == 'Occupancy':
            return self.status not in STATUS.ONLINE
        return super().is_on


class Door(BinarySensor):
    tags = ('door_contact',)



"""Abode lock device."""

from . import base
from . import status as STATUS


class Lock(base.Device):
    """Class to represent a door lock."""

    tags = ('door_lock',)

    def lock(self) -> None:
        """Lock the device."""
        self.set_status(int(STATUS.Lock.CLOSED))
        self._state['status'] = STATUS.Lock.CLOSED

    def unlock(self) -> None:
        """Unlock the device."""
        self.set_status(int(STATUS.Lock.OPEN))
        self._state['status'] = STATUS.Lock.OPEN

    @property
    def is_locked(self):
        """
        Get locked state.

        Err on side of caution, assume if lock isn't closed then it's open.
        """
        return self.status == STATUS.Lock.CLOSED



"""
Expose platform paths for the application.

>>> hasattr(paths, 'user_data')
True
"""

from __future__ import annotations

from collections.abc import Mapping

import pathlib
import platformdirs


class PlatformDirs(platformdirs.PlatformDirs):
    """
    Augment PlatformDirs to ensure the data path exists.

    >>> dirs = PlatformDirs(appname='Abode', appauthor=False)
    >>> dirs.override(user_data=getfixture('tmp_path') / 'data' / 'dir')
    >>> assert dirs.user_data.is_dir()
    """

    @property
    def user_data(self):
        self.user_data_path.mkdir(parents=True, exist_ok=True)
        return self.user_data_path

    @property
    def user_data_path(self):
        return vars(self).get('user_data_path') or super().user_data_path

    def override(self, **kwargs: Mapping[str, pathlib.Path]):
        """
        Override the default _path variable.

        >>> dirs = PlatformDirs(appname='Abode', appauthor=False)
        >>> dirs.override(user_data=getfixture('tmp_path') / 'foo')
        >>> 'foo' in str(dirs.user_data)
        True
        """
        vars(self).update({name + '_path': path for name, path in kwargs.items()})


paths = PlatformDirs(appname='Abode', appauthor=False)



"""Test the Abode device classes."""

from jaraco.abode.helpers import urls

from .mock import login as LOGIN
from .mock import logout as LOGOUT
from .mock import oauth_claims as OAUTH_CLAIMS
from .mock import panel as PANEL
from .mock.devices import lm as LM


class TestLM:
    """Test the sensor class/LM."""

    def test_cover_lm_properties(self, m):
        """Tests that sensor/LM devices properties work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=LM.device(lux='14 lx', humidity='34 %'))

        # Logout to reset everything
        self.client.logout()

        # Get our power switch
        device = self.client.get_device(LM.DEVICE_ID)

        # Test our device
        assert device is not None
        assert device.status == '72 °F'
        assert not device.battery_low
        assert not device.no_response
        assert device.has_temp
        assert device.has_humidity
        assert device.has_lux
        assert device.temp == 72
        assert device.temp_unit == '°F'
        assert device.humidity == 34
        assert device.humidity_unit == '%'
        assert device.lux == 14
        assert device.lux_unit == 'lux'

        # Set up our direct device get url
        device_url = urls.DEVICE.format(id=LM.DEVICE_ID)

        # Change device properties
        m.get(
            device_url,
            json=LM.device(
                status='12 °C',
                temp='12 °C',
                lux='100 lx',
                humidity='100 %',
                low_battery=True,
                no_response=True,
            ),
        )

        # Refesh device and test changes
        device.refresh()

        assert device.status == '12 °C'
        assert device.battery_low
        assert device.no_response
        assert device.has_temp
        assert device.has_humidity
        assert device.has_lux
        assert device.temp == 12
        assert device.temp_unit == '°C'
        assert device.humidity == 100
        assert device.humidity_unit == '%'
        assert device.lux == 100
        assert device.lux_unit == 'lux'

    def test_lm_float_units(self, m):
        """Tests that sensor/LM devices properties work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(
            urls.DEVICES,
            json=LM.device(
                status='72.23 °F',
                temp='72.23 °F',
                lux='14.11 lx',
                humidity='34.38 %',
            ),
        )

        # Logout to reset everything
        self.client.logout()

        # Get our power switch
        device = self.client.get_device(LM.DEVICE_ID)

        # Test our device
        assert device is not None
        assert device.status == '72.23 °F'
        assert not device.battery_low
        assert not device.no_response
        assert device.has_temp
        assert device.has_humidity
        assert device.has_lux
        assert device.temp == 72.23
        assert device.temp_unit == '°F'
        assert device.humidity == 34.38
        assert device.humidity_unit == '%'
        assert device.lux == 14.11
        assert device.lux_unit == 'lux'

    def test_lm_temp_only(self, m):
        """Tests that sensor/LM devices properties work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(
            urls.DEVICES,
            json=LM.device(status='72 °F', temp='72 °F', lux='', humidity=''),
        )

        # Logout to reset everything
        self.client.logout()

        # Get our power switch
        device = self.client.get_device(LM.DEVICE_ID)

        # Test our device
        assert device is not None
        assert device.status == '72 °F'
        assert device.has_temp
        assert not device.has_humidity
        assert not device.has_lux
        assert device.temp == 72
        assert device.temp_unit == '°F'
        assert device.humidity is None
        assert device.humidity_unit is None
        assert device.lux is None
        assert device.lux_unit is None



import urllib.parse

import pytest

import jaraco.abode


@pytest.fixture(autouse=True)
def instance_client(request):
    if request.instance is None:
        return

    request.instance.client = jaraco.abode.Client(
        username='foobar',
        password='deadbeef',
    )


def wrap_mock_register_uri(mocker):
    """
    Allow path to omit the leading /
    """
    orig = mocker.register_uri

    def register_uri(method, url, *args, **kwargs):
        if not urllib.parse.urlparse(url).path.startswith('/'):
            url = '/' + url
        return orig(method, url, *args, **kwargs)

    mocker.register_uri = register_uri
    return mocker


@pytest.fixture
def m(requests_mock):
    return wrap_mock_register_uri(requests_mock)


@pytest.fixture(autouse=True)
def app_paths(tmp_path, monkeypatch):
    """
    Redirect app dirs to temporary paths.
    """

    class Paths:
        user_data_path = tmp_path / 'user data'

        @property
        def user_data(self):
            self.user_data_path.mkdir(exist_ok=True)
            return self.user_data_path

    monkeypatch.setattr(jaraco.abode.config, 'paths', Paths())


@pytest.fixture
def all_devices(m):
    # Set up URLs
    import tests.mock.devices.door_contact
    import tests.mock.devices.door_lock
    import tests.mock.devices.glass
    import tests.mock.devices.ir_camera
    import tests.mock.devices.keypad
    import tests.mock.devices.pir
    import tests.mock.devices.power_switch_meter
    import tests.mock.devices.power_switch_sensor
    import tests.mock.devices.remote_controller
    import tests.mock.devices.secure_barrier
    import tests.mock.devices.siren
    import tests.mock.devices.status_display
    import tests.mock.devices.water_sensor
    import tests.mock.login
    import tests.mock.logout
    import tests.mock.oauth_claims
    import tests.mock.panel

    m.post(jaraco.abode.helpers.urls.LOGIN, json=tests.mock.login.post_response_ok())
    m.get(
        jaraco.abode.helpers.urls.OAUTH_TOKEN,
        json=tests.mock.oauth_claims.get_response_ok(),
    )
    m.post(jaraco.abode.helpers.urls.LOGOUT, json=tests.mock.logout.post_response_ok())
    m.get(
        jaraco.abode.helpers.urls.PANEL,
        json=tests.mock.panel.get_response_ok(mode='standby'),
    )

    # Create all devices
    all_devices = [
        tests.mock.devices.door_contact.device(),
        tests.mock.devices.door_lock.device(),
        tests.mock.devices.glass.device(),
        tests.mock.devices.ir_camera.device(),
        tests.mock.devices.keypad.device(),
        tests.mock.devices.pir.device(),
        tests.mock.devices.power_switch_meter.device(),
        tests.mock.devices.power_switch_sensor.device(),
        tests.mock.devices.remote_controller.device(),
        tests.mock.devices.secure_barrier.device(),
        tests.mock.devices.siren.device(),
        tests.mock.devices.status_display.device(),
        tests.mock.devices.water_sensor.device(),
    ]

    m.get(jaraco.abode.helpers.urls.DEVICES, json=all_devices)



"""Small SocketIO client via Websockets."""

import collections
import contextlib
import datetime
import itertools
import json
import logging
import random
import threading
import urllib.parse

from lomond import WebSocket, events
from lomond.errors import WebSocketError
from lomond.persist import persist

import jaraco.collections

from .exceptions import SocketIOException
from .helpers import errors as ERRORS

log = logging.getLogger(__name__)


class EngineIO:
    codes = jaraco.collections.BijectiveMap(
        open=0,
        close=1,
        ping=2,
        pong=3,
        message=4,
    )


class BackoffIntervals:
    """
    >>> bi = BackoffIntervals()
    >>> intervals = list(itertools.islice(bi, 10))
    >>> all(
    ...     BackoffIntervals.min_wait <= interval <= BackoffIntervals.max_wait
    ...     for interval in intervals)
    True
    """

    min_wait = 5
    max_wait = 30
    diff = max_wait - min_wait

    def __init__(self):
        self.reset(1)

    def __iter__(self):
        return self

    def __next__(self):
        attempt = next(self.attempts)
        return self.min_wait + random.random() * min(self.diff, 2**attempt)

    def reset(self, *args):
        self.attempts = itertools.count(*args)


def find_json_list(text):
    r"""
    >>> find_json_list('["foo",\n\t"bar"]')
    ['foo', 'bar']
    >>> find_json_list('{"abc": ["123"]}')
    ['123']
    >>> find_json_list('{"abc": null}')
    Traceback (most recent call last):
    ...
    ValueError: ...
    >>> find_json_list('some text string')
    Traceback (most recent call last):
    ...
    ValueError: ...
    """
    l_bracket = text.find("[")
    r_bracket = text.rfind("]")

    if l_bracket == -1 or r_bracket == -1:
        raise ValueError("No list found", text)

    return json.loads(text[l_bracket : r_bracket + 1])


class SocketIO:
    """Class for using websockets to talk to a SocketIO server."""

    codes = jaraco.collections.BijectiveMap(
        connect=0,
        disconnect=1,
        event=2,
        error=4,
    )

    def __init__(self, url, cookie=None, origin=None):
        params = dict(EIO=3, transport='websocket')
        self._url = url + '?' + urllib.parse.urlencode(params)

        self._cookie = cookie
        self._origin = origin

        self._thread = None
        self._websocket = None
        self._exit_event = None
        self._running = False

        self._websocket_connected = False
        self._engineio_connected = False
        self._socketio_connected = False

        self._last_ping_time = datetime.datetime.min
        self._last_packet_time = datetime.datetime.min

        self._callbacks = collections.defaultdict(list)

    def set_origin(self, origin=None):
        """Set the Origin header."""
        self._origin = origin

    def set_cookie(self, cookie=None):
        """Set the Cookie header."""
        self._cookie = cookie

    def on(self, event_name, callback):
        """Register callback for a SocketIO event."""
        log.debug("Adding callback for event name: %s", event_name)
        self._callbacks[event_name].append(callback)

    def start(self):
        """Start a thread to handle SocketIO notifications."""
        if self._thread:
            return

        log.info("Starting SocketIO thread...")

        self._thread = threading.Thread(
            target=self._run, name='SocketIOThread', daemon=True
        )
        self._thread.start()

    def stop(self):
        """Tell the SocketIO thread to terminate."""
        if not self._thread:
            return

        log.info("Stopping SocketIO thread...")

        self._running = False

        if self._exit_event:
            self._exit_event.set()

        self._thread.join()

    def _run(self):
        self._running = True

        intervals = BackoffIntervals()

        while self._running:
            log.info("Attempting to connect to SocketIO server...")

            try:
                self._step(intervals)
            except SocketIOException as exc:
                log.warning("SocketIO Error: %s", exc.details)
            except WebSocketError as exc:
                log.warning("Websocket Error: %s", exc)

            if not self._running:
                break

            interval = next(intervals)
            log.info("Waiting %f seconds before reconnecting...", interval)
            if self._exit_event.wait(interval):
                break

        self._handle_event('stopped', None)

    def _add_header(self, name, value):
        if value is None:
            return
        self._websocket.add_header(name.encode(), value.encode())

    def _step(self, intervals):
        self._handle_event('started')

        self._websocket = WebSocket(self._url)
        self._exit_event = threading.Event()

        self._add_header('Cookie', self._cookie)
        self._add_header('Origin', self._origin)

        for event in persist(
            self._websocket, ping_rate=0, poll=5.0, exit_event=self._exit_event
        ):
            if isinstance(event, events.Connected):
                intervals.reset()

            name = event.__class__.__name__.lower()
            with contextlib.suppress(AttributeError):
                handler = getattr(self, f'_on_websocket_{name}')
                handler(event)

            if self._running is False:
                self._websocket.close()

    def _on_websocket_connected(self, _event):
        self._websocket_connected = True
        log.info("Websocket Connected")
        self._handle_event('connected')

    def _on_websocket_disconnected(self, _event):
        self._websocket_connected = False
        self._engineio_connected = False
        self._socketio_connected = False

        log.info("Websocket Disconnected")
        self._handle_event('disconnected')

    def _on_websocket_poll(self, _event):
        last_packet_delta = datetime.datetime.now() - self._last_packet_time

        if self._engineio_connected and last_packet_delta > self._ping_timeout:
            log.warning("SocketIO Server Ping Timeout")
            self._websocket.close()
            return

        last_ping_delta = datetime.datetime.now() - self._last_ping_time

        if self._engineio_connected and last_ping_delta >= self._ping_interval:
            self._websocket.send_text(str(EngineIO.codes['ping']))
            self._last_ping_time = datetime.datetime.now()
            log.debug("Client Ping")
            self._handle_event('ping')

        self._handle_event('poll')

    def _on_websocket_text(self, _event):
        self._last_packet_time = datetime.datetime.now()

        log.debug("Received: %s", _event.text)

        code = int(_event.text[:1])
        message = _event.text[1:]

        try:
            name = EngineIO.codes[code]
            handler = getattr(self, f'_on_engineio_{name}')
            handler(message)
        except KeyError:
            log.debug("Ignoring unrecognized EngineIO packet")

    def _on_websocket_backoff(self, _event):
        return

    def _on_engineio_open(self, message):
        packet = json.loads(message)

        self._ping_interval = datetime.timedelta(milliseconds=packet['pingInterval'])
        log.debug("Set ping interval to %s", self._ping_interval)

        self._ping_timeout = datetime.timedelta(milliseconds=packet['pingTimeout'])
        log.debug("Set ping timeout to %s", self._ping_timeout)

        self._engineio_connected = True
        log.debug("EngineIO Connected")

    def _on_engineio_close(self, message):
        self._engineio_connected = False
        log.debug("EngineIO Disconnected")
        self._websocket.close()

    def _on_engineio_pong(self, message):
        log.debug("Server Pong")
        self._handle_event('pong')

    def _on_engineio_message(self, message):
        code = int(message[:1])
        data = message[1:]

        try:
            name = SocketIO.codes[code]
            handler = getattr(self, f'_on_socketio_{name}')
            handler(data)
        except KeyError:
            log.debug("Ignoring SocketIO message: %s", message)

    def _on_socketio_connected(self):
        self._socketio_connected = True
        log.debug("SocketIO Connected")

    def _on_socketio_disconnected(self):
        self._socketio_connected = False
        log.debug("SocketIO Disconnected")
        self._websocket.close()

    def _on_socketio_error(self, _message_data):
        self._handle_event('error', _message_data)
        raise SocketIOException(ERRORS.SOCKETIO_ERROR, details=_message_data)

    def _on_socketio_event(self, _message_data):
        try:
            json_data = find_json_list(_message_data)
        except ValueError:
            log.warning("Unable to find event [data]: %s", _message_data)
            return
        self._handle_event('event', _message_data)
        self._handle_event(json_data[0], json_data[1:])

    def _handle_event(self, event_name, *args):
        for callback in self._callbacks[event_name]:
            try:
                callback(*args)
            except Exception as exc:
                log.exception(
                    "Captured exception during SocketIO event callback: %s", exc
                )



"""Mock Abode IP Camera Device."""

import jaraco.abode.devices.status as STATUS

DEVICE_ID = 'ZB:00000305'
CONTROL_URL = 'api/v1/cams/' + DEVICE_ID + '/record'
CONTROL_URL_SNAPSHOT = 'api/v1/cams/' + DEVICE_ID + '/capture'


def device(
    devid=DEVICE_ID,
    status=STATUS.ONLINE,
    low_battery=False,
    no_response=False,
    privacy=1,
):
    """IP camera mock device."""
    return dict(
        id=devid,
        type_tag='device_type.ipcam',
        type='IP Cam',
        name='Living Room Camera',
        area='1',
        zone='1',
        sort_order='',
        is_window='',
        bypass='0',
        schar_24hr='1',
        sresp_24hr='5',
        sresp_mode_0='0',
        sresp_entry_0='0',
        sresp_exit_0='0',
        group_name='Streaming Camera',
        group_id='397974',
        default_group_id='1',
        sort_id='10000',
        sresp_mode_1='0',
        sresp_entry_1='0',
        sresp_exit_1='0',
        sresp_mode_2='0',
        sresp_entry_2='0',
        sresp_exit_2='0',
        sresp_mode_3='0',
        uuid='123456789',
        sresp_entry_3='0',
        sresp_exit_3='0',
        sresp_mode_4='0',
        sresp_entry_4='0',
        sresp_exit_4='0',
        version='1.0.2.22G_6.8E_homekit_2.0.9_s2 ABODE oz',
        origin='abode',
        has_subscription=None,
        onboard='1',
        s2_grnt_keys='',
        s2_dsk='',
        s2_propty='',
        s2_keys_valid='',
        zwave_secure_protocol='',
        control_url='api/v1/cams/ZB:00000305/record',
        deep_link=None,
        status_color='#5cb85c',
        faults={
            'low_battery': int(low_battery),
            'tempered': 0,
            'supervision': 0,
            'out_of_order': 0,
            'no_response': int(no_response),
            'jammed': 0,
            'zwave_fault': 0,
        },
        status=status,
        status_display='Online',
        statuses=[],
        status_ex='',
        actions=[
            {'label': 'Capture Video', 'value': 'a=1&z=1&req=vid;'},
            {'label': 'Turn off Live Video', 'value': 'a=1&z=1&privacy=on;'},
            {'label': 'Turn on Live Video', 'value': 'a=1&z=1&privacy=off;'},
        ],
        status_icons=[],
        icon='assets/icons/streaming-camaera-new.svg',
        control_url_snapshot=CONTROL_URL_SNAPSHOT,
        ptt_supported=True,
        is_new_camera=1,
        stream_quality=3,
        camera_mac='AB:CD:EF:GF:HI',
        privacy=str(privacy),
        enable_audio='1',
        alarm_video='25',
        pre_alarm_video='5',
        mic_volume='75',
        speaker_volume='75',
        mic_default_volume=40,
        speaker_default_volume=46,
        bandwidth={
            'slider_labels': [
                {'name': 'High', 'value': 3},
                {'name': 'Medium', 'value': 2},
                {'name': 'Low', 'value': 1},
            ],
            'min': 1,
            'max': 3,
            'step': 1,
        },
        volume={'min': 0, 'max': 100, 'step': 1},
        video_flip='0',
        hframe='1080P',
    )


def get_capture_timeout():
    """Mock timeout response."""
    return dict(
        code=600, message='Image Capture request has timed out.', title='', detail=None
    )


FILE_PATH_ID = 'ZB00000305'
FILE_PATH = 'api/storage/' + FILE_PATH_ID + '/2020-01-26/173238/0.jpg'

LOCATION_HEADER = (
    'https://www.google.com/images/branding/googlelogo/'
    + '1x/googlelogo_color_272x92dp.png'
)


def timeline_event(devid=DEVICE_ID, event_code='5001', file_path=FILE_PATH):
    """Camera Timeline Event Mockup."""
    return dict(
        mac='B0:C5:CZ:54:12:9A',
        id=devid,
        xml=None,
        date='01/26/2020',
        time='05:32 PM',
        event_utc='1580088758',
        event_cid='',
        event_code=event_code,
        device_id='ZB:00000305',
        device_type_id='69',
        device_type='IP Cam',
        timeline_ha_device=None,
        d_name='Living Room Camera',
        delete_by_user=None,
        pin_code_user=' ',
        file_del_at='',
        nest_has_motion=None,
        nest_has_sound=None,
        nest_has_person=None,
        neaz=None,
        hasFaults='0',
        file_path=file_path,
        deep_link=None,
        file_name='48755_b0c5ca37894b_2020-01-26_173238_0-M2+56431.jpg',
        file_size='197207',
        file_count='1',
        file_is_del='0',
        event_type='Image Capture',
        severity='6',
        pos='l',
        color='#40bbea',
        is_alarm='0',
        triggered_by_str=None,
        ha_type=None,
        ha_device_name=None,
        ha_location=None,
        ha_mobile=None,
        ha_cond=None,
        h_location=None,
        ha_trigger=None,
        icon='assets/email/motion-camera.png',
        user_id='95244',
        user_name='Shred',
        mobile_name='',
        parent_tid='',
        app_type='WebApp',
        viewed_by_uid=None,
        verified_by_tid=None,
        la_applied_by=None,
        la_event_type=None,
        la_culprit_mobiles=None,
        la_executed=None,
        la_applied_at=None,
        device_name='Living Room Camera',
        event_name='Living Room Camera Image Capture',
        event_by='by Shred using WebApp',
        file_delete_by='',
    )



"""Mock Abode Door Contact Device."""

import jaraco.abode.devices.status as STATUS

DEVICE_ID = 'RF:00000003'


def device(
    devid=DEVICE_ID,
    status=STATUS.CLOSED,
    low_battery=False,
    no_response=False,
    window=False,
):
    """Door contact mock device."""
    return dict(
        id=devid,
        type_tag='device_type.door_contact',
        type='Door Contact',
        name='Back Door',
        area='1',
        zone='2',
        sort_order=None,
        is_window=int(window),
        bypass='0',
        schar_24hr='0',
        sresp_mode_0='3',
        sresp_entry_0='3',
        sresp_exit_0='0',
        sresp_mode_1='1',
        sresp_entry_1='1',
        sresp_exit_1='0',
        sresp_mode_2='1',
        sresp_entry_2='1',
        sresp_exit_2='0',
        sresp_mode_3='1',
        sresp_entry_3='1',
        sresp_exit_3='0',
        version='',
        origin='abode',
        control_url='',
        deep_link=None,
        status_color='#5cb85c',
        faults={
            'low_battery': int(low_battery),
            'tempered': 0,
            'supervision': 0,
            'out_of_order': 0,
            'no_response': int(no_response),
        },
        status=status,
        statuses={'hvac_mode': None},
        status_ex='',
        actions=[],
        status_icons={
            'Open': 'assets/icons/door-open-red.svg',
            'Closed': 'assets/icons/door-closed-green.svg',
        },
        sresp_trigger='0',
        sresp_restore='0',
        icon='assets/icons/doorsensor-a.svg',
    )



"""Test the Abode device classes."""

import pytest

import jaraco.abode
import jaraco.abode.devices.status as STATUS
from jaraco.abode.helpers import urls

from .mock import devices as DEVICES
from .mock import login as LOGIN
from .mock import logout as LOGOUT
from .mock import oauth_claims as OAUTH_CLAIMS
from .mock import panel as PANEL
from .mock.devices import door_lock as DOOR_LOCK


class TestDoorLock:
    """Test the door lock."""

    def test_lock_device_properties(self, m):
        """Tests that lock devices properties work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=DOOR_LOCK.device())

        # Logout to reset everything
        self.client.logout()

        # Get our lock
        device = self.client.get_device(DOOR_LOCK.DEVICE_ID)

        # Test our device
        assert device is not None
        assert device.status == STATUS.Lock.CLOSED
        assert not device.battery_low
        assert not device.no_response
        assert device.is_locked

        # Set up our direct device get url
        device_url = urls.DEVICE.format(id=DOOR_LOCK.DEVICE_ID)

        # Change device properties
        m.get(
            device_url,
            json=DOOR_LOCK.device(
                status=STATUS.Lock.OPEN,
                low_battery=True,
                no_response=True,
            ),
        )

        # Refesh device and test changes
        device.refresh()

        assert device.status == STATUS.Lock.OPEN
        assert device.battery_low
        assert device.no_response
        assert not device.is_locked

    def test_lock_device_mode_changes(self, m):
        """Tests that lock device changes work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=DOOR_LOCK.device())

        # Logout to reset everything
        self.client.logout()

        # Get our power switch
        device = self.client.get_device(DOOR_LOCK.DEVICE_ID)

        # Test that we have our device
        assert device is not None
        assert device.status == STATUS.Lock.CLOSED
        assert device.is_locked

        # Set up control url response
        control_url = urls.BASE + DOOR_LOCK.CONTROL_URL
        m.put(
            control_url,
            json=DEVICES.status_put_response_ok(
                devid=DOOR_LOCK.DEVICE_ID, status=int(STATUS.Lock.OPEN)
            ),
        )

        # Change the mode to "on"
        device.unlock()
        assert device.status == STATUS.Lock.OPEN
        assert not device.is_locked

        # Change response
        m.put(
            control_url,
            json=DEVICES.status_put_response_ok(
                devid=DOOR_LOCK.DEVICE_ID, status=int(STATUS.Lock.CLOSED)
            ),
        )

        # Change the mode to "off"
        device.lock()
        assert device.status == STATUS.Lock.CLOSED
        assert device.is_locked

        # Test that an invalid status response throws exception
        m.put(
            control_url,
            json=DEVICES.status_put_response_ok(
                devid=DOOR_LOCK.DEVICE_ID, status=int(STATUS.Lock.CLOSED)
            ),
        )

        with pytest.raises(jaraco.abode.Exception):
            device.unlock()



import functools
import importlib

import importlib_resources as resources
from bx_py_utils.compat import removesuffix


@functools.lru_cache
def import_all():
    """Import all modules from this package."""
    device_mods = (
        removesuffix(mod.name, '.py')
        for mod in resources.files(__package__).iterdir()
        if mod.name != '__init__.py'
    )
    for mod in device_mods:
        importlib.import_module(f'.{mod}', __package__)



"""Abode cloud push events."""

import collections
import http.cookiejar
import logging

import jaraco
from jaraco.itertools import always_iterable

from . import socketio as sio
from ._itertools import opt_single, single
from .devices.alarm import Alarm
from .devices.base import Device
from .helpers import errors as ERROR
from .helpers import timeline as TIMELINE
from .helpers import urls

log = logging.getLogger(__name__)

SOCKETIO_URL = 'wss://my.goabode.com/socket.io/'


def _cookie_string(cookies: http.cookiejar.CookieJar):
    """
    >>> jar = http.cookiejar.CookieJar()
    >>> from test.test_http_cookiejar import interact_netscape
    >>> _ = interact_netscape(jar, 'http://any/', 'foo=bar', 'bing=baz')
    >>> _cookie_string(jar) in ['foo=bar; bing=baz', 'bing=baz; foo=bar']
    True
    """
    return "; ".join(f"{cookie.name}={cookie.value}" for cookie in cookies)


class EventController:
    """Subscribes to events."""

    def __init__(self, client, url=SOCKETIO_URL):
        self._client = client
        self._thread = None
        self._running = False
        self._connected = False

        # Setup callback dicts
        self._connection_status_callbacks = collections.defaultdict(list)
        self._device_callbacks = collections.defaultdict(list)
        self._event_callbacks = collections.defaultdict(list)
        self._timeline_callbacks = collections.defaultdict(list)

        # Setup SocketIO
        self._socketio = sio.SocketIO(url=url, origin=urls.BASE)

        # Setup SocketIO Callbacks
        self._socketio.on('started', self._on_socket_started)
        self._socketio.on('connected', self._on_socket_connected)
        self._socketio.on('disconnected', self._on_socket_disconnected)
        self._socketio.on('com.goabode.device.update', self._on_device_update)
        self._socketio.on('com.goabode.gateway.mode', self._on_mode_change)
        self._socketio.on('com.goabode.gateway.timeline', self._on_timeline_update)
        self._socketio.on('com.goabode.automation', self._on_automation_update)

    def start(self):
        """Start a thread to handle Abode SocketIO notifications."""
        self._socketio.start()

    def stop(self):
        """Tell the subscription thread to terminate - will block."""
        self._socketio.stop()

    def add_connection_status_callback(self, unique_id, callback):
        """Register callback for Abode server connection status."""
        if not unique_id:
            return False

        log.debug("Subscribing to Abode connection updates for: %s", unique_id)

        self._connection_status_callbacks[unique_id].append(callback)

        return True

    def remove_connection_status_callback(self, unique_id):
        """Unregister connection status callbacks."""
        if not unique_id:
            return False

        log.debug("Unsubscribing from Abode connection updates for : %s", unique_id)

        self._connection_status_callbacks[unique_id].clear()

        return True

    def add_device_callback(self, devices, callback):
        """Register a device callback."""
        if not devices:
            return False

        for device in always_iterable(devices):
            # Device may be a device_id
            device_id = device

            # If they gave us an actual device, get that devices ID
            if isinstance(device, Device):
                device_id = device.id

            # Validate the device is valid
            if not self._client.get_device(device_id):
                raise jaraco.abode.Exception(ERROR.EVENT_DEVICE_INVALID)

            log.debug("Subscribing to updates for device_id: %s", device_id)

            self._device_callbacks[device_id].append(callback)

        return True

    def remove_all_device_callbacks(self, devices):
        """Unregister all callbacks for a device."""
        if not devices:
            return False

        for device in always_iterable(devices):
            device_id = device

            if isinstance(device, Device):
                device_id = device.id

            if not self._client.get_device(device_id):
                raise jaraco.abode.Exception(ERROR.EVENT_DEVICE_INVALID)

            if device_id not in self._device_callbacks:
                return False

            log.debug("Unsubscribing from all updates for device_id: %s", device_id)

            self._device_callbacks[device_id].clear()

        return True

    def add_event_callback(self, event_groups, callback):
        """Register callback for a group of timeline events."""
        if not event_groups:
            return False

        for event_group in always_iterable(event_groups):
            if event_group not in TIMELINE.Groups.ALL:
                raise jaraco.abode.Exception(ERROR.EVENT_GROUP_INVALID)

            log.debug("Subscribing to event group: %s", event_group)

            self._event_callbacks[event_group].append(callback)

        return True

    def add_timeline_callback(self, timeline_events, callback):
        """Register a callback for a specific timeline event."""
        if not timeline_events:
            return False

        for timeline_event in always_iterable(timeline_events):
            if not isinstance(timeline_event, dict):
                raise jaraco.abode.Exception(ERROR.EVENT_CODE_MISSING)

            event_code = timeline_event.get('event_code')

            if not event_code:
                raise jaraco.abode.Exception(ERROR.EVENT_CODE_MISSING)

            log.debug("Subscribing to timeline event: %s", timeline_event)

            self._timeline_callbacks[event_code].append(callback)

        return True

    @property
    def connected(self):
        """Get the Abode connection status."""
        return self._connected

    @property
    def socketio(self):
        """Get the SocketIO instance."""
        return self._socketio

    def _on_socket_started(self):
        """Socket IO startup callback."""
        self._socketio.set_cookie(_cookie_string(self._client._get_session().cookies))

    def _on_socket_connected(self):
        """Socket IO connected callback."""
        self._connected = True

        try:
            self._client.refresh()
        except Exception as exc:
            log.warning("Captured exception during Abode refresh: %s", exc)
        finally:
            # Callbacks should still execute even if refresh fails (Abode
            # server issues) so that the entity availability in Home Assistant
            # is updated since we are in fact connected to the web socket.
            for callbacks in self._connection_status_callbacks.values():
                for callback in callbacks:
                    _execute_callback(callback)

    def _on_socket_disconnected(self):
        """Socket IO disconnected callback."""
        self._connected = False

        for callbacks in self._connection_status_callbacks.values():
            for callback in callbacks:
                _execute_callback(callback)

    def _on_device_update(self, devid):
        """Device callback from Abode SocketIO server."""
        devid = opt_single(devid)

        if devid is None:
            log.warning("Device update with no device id.")
            return

        log.debug("Device update event for device ID: %s", devid)

        device = self._client.get_device(devid, True)

        if not device:
            log.debug("Got device update for unknown device: %s", devid)
            return

        for callback in self._device_callbacks[device.id]:
            _execute_callback(callback, device)

    def _on_mode_change(self, mode):
        """Mode change broadcast from Abode SocketIO server."""
        mode = opt_single(mode)

        if mode is None:
            log.warning("Mode change event with no mode.")
            return

        if not mode or mode.lower() not in Alarm.all_modes:
            log.warning("Mode change event with unknown mode: %s", mode)
            return

        log.debug("Alarm mode change event to: %s", mode)

        # We're just going to convert it to an Alarm device
        alarm_device = self._client.get_alarm(refresh=True)

        # At the time of development, refreshing after mode change notification
        # didn't seem to get the latest update immediately. As such, we will
        # force the mode status now to match the notification.
        alarm_device._state['mode']['area_1'] = mode

        for callback in self._device_callbacks[alarm_device.id]:
            _execute_callback(callback, alarm_device)

    def _on_timeline_update(self, event):
        """Timeline update broadcast from Abode SocketIO server."""
        event = single(event)

        event_type = event.get('event_type')
        event_code = event.get('event_code')

        if not event_type or not event_code:
            log.warning("Invalid timeline update event: %s", event)
            return

        log.debug(
            "Timeline event received: %s - %s (%s)",
            event.get('event_name'),
            event_type,
            event_code,
        )

        # Compress our callbacks into those that match this event_code
        # or ones registered to get callbacks for all events
        codes = (event_code, TIMELINE.ALL['event_code'])
        all_callbacks = [self._timeline_callbacks[code] for code in codes]

        for callbacks in all_callbacks:
            for callback in callbacks:
                _execute_callback(callback, event)

        # Attempt to map the event code to a group and callback
        event_group = TIMELINE.map_event_code(event_code)

        for callback in self._event_callbacks[event_group]:
            _execute_callback(callback, event)

    def _on_automation_update(self, event):
        """Automation update broadcast from Abode SocketIO server."""
        event_group = TIMELINE.Groups.AUTOMATION_EDIT

        event = single(event)

        for callback in self._event_callbacks[event_group]:
            _execute_callback(callback, event)


def _execute_callback(callback, *args, **kwargs):
    # Callback with some data, capturing any exceptions to prevent chaos
    try:
        callback(*args, **kwargs)
    except Exception as exc:
        log.warning("Captured exception during callback: %s", exc)



import more_itertools

import jaraco.functools
import jaraco.itertools

single = jaraco.functools.compose(more_itertools.one, jaraco.itertools.always_iterable)

opt_single = jaraco.functools.compose(
    more_itertools.only, jaraco.itertools.always_iterable
)



"""Mock Abode Power Switch Sensor Device."""

import jaraco.abode.devices.status as STATUS
from jaraco.abode.devices.light import ColorMode
from jaraco.abode.helpers import urls

DEVICE_ID = 'ZB:00000106'
DEVICE_UUID = 'abcd33455232fff31232'
CONTROL_URL = 'api/v1/control/light/' + DEVICE_ID
INTEGRATIONS_URL = urls.INTEGRATIONS + DEVICE_UUID


def color_temp_post_response_ok(devid, color_temp):
    """Return color temp change response json."""
    return dict(idForPanel=devid, colorTemperature=int(color_temp))


def color_post_response_ok(devid, hue, saturation):
    """Return color change response json."""
    return dict(idForPanel=devid, hue=int(hue), saturation=int(saturation))


def device(
    devid=DEVICE_ID,
    uuid=DEVICE_UUID,
    status=STATUS.OFF,
    level=0,
    saturation=57,
    hue=60,
    color_temp=6536,
    color_mode=ColorMode.off,
    low_battery=False,
    no_response=False,
):
    """Hue mock device."""
    return dict(
        id=devid,
        type_tag='device_type.hue',
        type='RGB Dimmer',
        name='Overhead Light',
        area='1',
        zone='30',
        sort_order='',
        is_window='',
        bypass='0',
        schar_24hr='0',
        sresp_24hr='0',
        sresp_mode_0='0',
        sresp_entry_0='0',
        sresp_exit_0='0',
        group_name='Ungrouped',
        group_id='1',
        default_group_id='1',
        sort_id='10000',
        sresp_mode_1='0',
        sresp_entry_1='0',
        sresp_exit_1='0',
        sresp_mode_2='0',
        sresp_entry_2='0',
        sresp_exit_2='0',
        sresp_mode_3='0',
        uuid=DEVICE_UUID,
        sresp_entry_3='0',
        sresp_exit_3='0',
        sresp_mode_4='0',
        sresp_entry_4='0',
        sresp_exit_4='0',
        version='LST002',
        origin='abode',
        has_subscription=None,
        control_url=CONTROL_URL,
        deep_link=None,
        status_color='#5cb85c',
        faults={
            'low_battery': int(low_battery),
            'tempered': 0,
            'supervision': 0,
            'out_of_order': 0,
            'no_response': int(no_response),
            'jammed': 0,
            'zwave_fault': 0,
        },
        status=status,
        statuses={
            'saturation': int(saturation),
            'hue': int(hue),
            'level': str(int(level)),
            'switch': '1',
            'color_temp': int(color_temp),
            'color_mode': str(color_mode),
        },
        status_ex='',
        actions=[
            {'label': 'Switch off', 'value': 'a=1&z=30&sw=off;'},
            {'label': 'Switch on', 'value': 'a=1&z=30&sw=on;'},
            {'label': 'Toggle', 'value': 'a=1&z=30&sw=toggle;'},
            {'label': '0%', 'value': 'a=1&z=30&sw=0;'},
            {'label': '10%', 'value': 'a=1&z=30&sw=10;'},
            {'label': '20%', 'value': 'a=1&z=30&sw=20;'},
            {'label': '30%', 'value': 'a=1&z=30&sw=30;'},
            {'label': '40%', 'value': 'a=1&z=30&sw=40;'},
            {'label': '50%', 'value': 'a=1&z=30&sw=50;'},
            {'label': '60%', 'value': 'a=1&z=30&sw=60;'},
            {'label': '70%', 'value': 'a=1&z=30&sw=70;'},
            {'label': '80%', 'value': 'a=1&z=30&sw=80;'},
            {'label': '90%', 'value': 'a=1&z=30&sw=90;'},
            {'label': '100%', 'value': 'a=1&z=30&sw=99;'},
        ],
        status_icons=[],
        statusEx='37',
        icon='assets/icons/bulb-1.svg',
    )



import types

import jaraco.abode

from .helpers import errors as ERROR
from .helpers import urls

CAMERA_RESOLUTION = 'ircamera_resolution_t'
CAMERA_GRAYSCALE = 'ircamera_gray_t'
SILENCE_SOUNDS = 'beeper_mute'

PANEL_SETTINGS = [
    CAMERA_RESOLUTION,
    CAMERA_GRAYSCALE,
    SILENCE_SOUNDS,
]

ENTRY_DELAY_AWAY = 'away_entry_delay'
EXIT_DELAY_AWAY = 'away_exit_delay'
ENTRY_DELAY_HOME = 'home_entry_delay'
EXIT_DELAY_HOME = 'home_exit_delay'

AREA_SETTINGS = [
    ENTRY_DELAY_AWAY,
    EXIT_DELAY_AWAY,
    ENTRY_DELAY_HOME,
    EXIT_DELAY_HOME,
]

DOOR_CHIME = 'door_chime'
WARNING_BEEP = 'warning_beep'
ENTRY_BEEP_AWAY = 'entry_beep_away'
EXIT_BEEP_AWAY = 'exit_beep_away'
ENTRY_BEEP_HOME = 'entry_beep_home'
EXIT_BEEP_HOME = 'exit_beep_home'
CONFIRM_SOUND = 'confirm_snd'
ALARM_LENGTH = 'alarm_len'
FINAL_BEEPS = 'final_beep'

SOUND_SETTINGS = [
    DOOR_CHIME,
    WARNING_BEEP,
    ENTRY_BEEP_AWAY,
    EXIT_BEEP_AWAY,
    ENTRY_BEEP_HOME,
    EXIT_BEEP_HOME,
    CONFIRM_SOUND,
    ALARM_LENGTH,
    FINAL_BEEPS,
]

SIREN_ENTRY_EXIT_SOUNDS = "entry"
SIREN_TAMPER_SOUNDS = "tamper"
SIREN_CONFIRM_SOUNDS = "confirm"

SIREN_SETTINGS = [
    SIREN_ENTRY_EXIT_SOUNDS,
    SIREN_TAMPER_SOUNDS,
    SIREN_CONFIRM_SOUNDS,
]

CAMERA_RES_320_240 = '0'
CAMERA_RES_640_480 = '2'

ALL_CAMERA_RES = [CAMERA_RES_320_240, CAMERA_RES_640_480]

DISABLE = '0'
ENABLE = '1'

DISABLE_ENABLE = [DISABLE, ENABLE]

ENTRY_EXIT_DELAY_DISABLE = '0'
ENTRY_EXIT_DELAY_10SEC = '10'
ENTRY_EXIT_DELAY_20SEC = '20'
ENTRY_EXIT_DELAY_30SEC = '30'
ENTRY_EXIT_DELAY_1MIN = '60'
ENTRY_EXIT_DELAY_2MIN = '120'
ENTRY_EXIT_DELAY_3MIN = '180'
ENTRY_EXIT_DELAY_4MIN = '240'

ALL_ENTRY_EXIT_DELAY = [
    ENTRY_EXIT_DELAY_DISABLE,
    ENTRY_EXIT_DELAY_10SEC,
    ENTRY_EXIT_DELAY_20SEC,
    ENTRY_EXIT_DELAY_30SEC,
    ENTRY_EXIT_DELAY_1MIN,
    ENTRY_EXIT_DELAY_2MIN,
    ENTRY_EXIT_DELAY_3MIN,
    ENTRY_EXIT_DELAY_4MIN,
]

VALID_SETTING_EXIT_AWAY = [
    ENTRY_EXIT_DELAY_30SEC,
    ENTRY_EXIT_DELAY_1MIN,
    ENTRY_EXIT_DELAY_2MIN,
    ENTRY_EXIT_DELAY_3MIN,
    ENTRY_EXIT_DELAY_4MIN,
]

SOUND_OFF = 'none'
SOUND_LOW = 'normal'
SOUND_HIGH = 'loud'

ALL_SOUND = [SOUND_OFF, SOUND_LOW, SOUND_HIGH]

VALID_SOUND_SETTINGS = [
    DOOR_CHIME,
    WARNING_BEEP,
    ENTRY_BEEP_AWAY,
    EXIT_BEEP_AWAY,
    ENTRY_BEEP_HOME,
    EXIT_BEEP_HOME,
    CONFIRM_SOUND,
]

ALARM_LENGTH_DISABLE = '0'
ALARM_LENGTH_1MIN = '60'
ALARM_LENGTH_2MIN = '120'
ALARM_LENGTH_3MIN = '180'
ALARM_LENGTH_4MIN = '240'
ALARM_LENGTH_5MIN = '300'
ALARM_LENGTH_6MIN = '360'
ALARM_LENGTH_7MIN = '420'
ALARM_LENGTH_8MIN = '480'
ALARM_LENGTH_9MIN = '540'
ALARM_LENGTH_10MIN = '600'
ALARM_LENGTH_11MIN = '660'
ALARM_LENGTH_12MIN = '720'
ALARM_LENGTH_13MIN = '780'
ALARM_LENGTH_14MIN = '840'
ALARM_LENGTH_15MIN = '900'

ALL_ALARM_LENGTH = [
    ALARM_LENGTH_DISABLE,
    ALARM_LENGTH_1MIN,
    ALARM_LENGTH_2MIN,
    ALARM_LENGTH_3MIN,
    ALARM_LENGTH_4MIN,
    ALARM_LENGTH_5MIN,
    ALARM_LENGTH_6MIN,
    ALARM_LENGTH_7MIN,
    ALARM_LENGTH_8MIN,
    ALARM_LENGTH_9MIN,
    ALARM_LENGTH_10MIN,
    ALARM_LENGTH_11MIN,
    ALARM_LENGTH_12MIN,
    ALARM_LENGTH_13MIN,
    ALARM_LENGTH_14MIN,
    ALARM_LENGTH_15MIN,
]

FINAL_BEEPS_DISABLE = '0'
FINAL_BEEPS_3SEC = '3'
FINAL_BEEPS_4SEC = '4'
FINAL_BEEPS_5SEC = '5'
FINAL_BEEPS_6SEC = '6'
FINAL_BEEPS_7SEC = '7'
FINAL_BEEPS_8SEC = '8'
FINAL_BEEPS_9SEC = '9'
FINAL_BEEPS_10SEC = '10'

ALL_FINAL_BEEPS = [
    FINAL_BEEPS_DISABLE,
    FINAL_BEEPS_3SEC,
    FINAL_BEEPS_4SEC,
    FINAL_BEEPS_5SEC,
    FINAL_BEEPS_6SEC,
    FINAL_BEEPS_7SEC,
    FINAL_BEEPS_8SEC,
    FINAL_BEEPS_9SEC,
    FINAL_BEEPS_10SEC,
]


class Setting(types.SimpleNamespace):
    @classmethod
    def load(cls, name, value, area):
        matches = (subcl for subcl in cls.__subclasses__() if name in subcl.names)
        try:
            (match,) = matches
        except ValueError as err:
            raise jaraco.abode.Exception(ERROR.INVALID_SETTING) from err

        ob = match(name=name, value=value, area=area)
        ob.validate()
        return ob


class Panel(Setting):
    names = PANEL_SETTINGS
    path = urls.SETTINGS

    def validate(self):
        if self.name == CAMERA_RESOLUTION and self.value not in ALL_CAMERA_RES:
            raise jaraco.abode.Exception(ERROR.INVALID_SETTING_VALUE)
        if (
            self.name in [CAMERA_GRAYSCALE, SILENCE_SOUNDS]
            and self.value not in DISABLE_ENABLE
        ):
            raise jaraco.abode.Exception(ERROR.INVALID_SETTING_VALUE)

    @property
    def data(self):
        return {self.name: self.value}


class Area(Setting):
    names = AREA_SETTINGS
    path = urls.AREAS

    def validate(self):
        # Exit delay has some specific limitations
        if self.name == EXIT_DELAY_AWAY and self.value not in VALID_SETTING_EXIT_AWAY:
            raise jaraco.abode.Exception(ERROR.INVALID_SETTING_VALUE)
        if self.value not in ALL_ENTRY_EXIT_DELAY:
            raise jaraco.abode.Exception(ERROR.INVALID_SETTING_VALUE)

    @property
    def data(self):
        return {'area': self.area, self.name: self.value}


class Sound(Setting):
    names = SOUND_SETTINGS
    path = urls.SOUNDS

    def validate(self):
        if self.name in VALID_SOUND_SETTINGS and self.value not in ALL_SOUND:
            raise jaraco.abode.Exception(ERROR.INVALID_SETTING_VALUE)
        if self.name == ALARM_LENGTH and self.value not in ALL_ALARM_LENGTH:
            raise jaraco.abode.Exception(ERROR.INVALID_SETTING_VALUE)
        if self.name == FINAL_BEEPS and self.value not in ALL_FINAL_BEEPS:
            raise jaraco.abode.Exception(ERROR.INVALID_SETTING_VALUE)

    @property
    def data(self):
        return {'area': self.area, self.name: self.value}


class Siren(Setting):
    names = SIREN_SETTINGS
    path = urls.SIREN

    def validate(self):
        if self.value not in DISABLE_ENABLE:
            raise jaraco.abode.Exception(ERROR.INVALID_SETTING_VALUE)

    @property
    def data(self):
        return {'action': self.name, 'option': self.value}



"""Mock Abode Power Switch Sensor Device."""

import jaraco.abode.devices.status as STATUS

DEVICE_ID = 'RF:00000008'


def device(devid=DEVICE_ID, status=STATUS.OFF, low_battery=False, no_response=False):
    """Water sensor mock device."""
    return dict(
        id=devid,
        type_tag='device_type.water_sensor',
        type='Water Sensor',
        name='Downstairs Bathroo',
        area='1',
        zone='26',
        sort_order='0',
        is_window='0',
        bypass='0',
        schar_24hr='1',
        sresp_mode_0='0',
        sresp_entry_0='0',
        sresp_exit_0='0',
        sresp_mode_1='0',
        sresp_entry_1='0',
        sresp_exit_1='0',
        sresp_mode_2='0',
        sresp_entry_2='0',
        sresp_exit_2='0',
        sresp_mode_3='0',
        sresp_entry_3='0',
        sresp_exit_3='0',
        capture_mode=None,
        origin='abode',
        control_url='',
        deep_link=None,
        status_color='#5cb85c',
        faults={
            'low_battery': int(low_battery),
            'tempered': 0,
            'supervision': 0,
            'out_of_order': 0,
            'no_response': int(no_response),
        },
        status=status,
        statuses={'hvac_mode': None},
        status_ex='',
        actions=[],
        status_icons=[],
        icon='assets/icons/water-value-shutoff.svg',
    )



"""Abode cover device."""

from . import status as STATUS
from .switch import Switch


class Cover(Switch):
    """Class to add cover functionality."""

    tags = ('secure_barrier',)

    def switch_on(self):
        """Turn the switch on."""
        self.set_status(int(STATUS.OPEN))
        self._state['status'] = STATUS.OPEN

    def switch_off(self):
        """Turn the switch off."""
        self.set_status(int(STATUS.CLOSED))
        self._state['status'] = STATUS.CLOSED

    def open_cover(self):
        """Open the cover."""
        return self.switch_on()

    def close_cover(self):
        """Close the cover."""
        return self.switch_off()

    @property
    def is_open(self):
        """Get if the cover is open."""
        return self.is_on

    @property
    def is_on(self):
        """
        Get cover state.

        Assume cover is open.
        """
        return self.status not in STATUS.CLOSED



"""Mock Abode Power Switch Sensor Device."""

import jaraco.abode.devices.status as STATUS

DEVICE_ID = 'ZW:00000113'
CONTROL_URL = 'api/v1/control/power_switch/' + DEVICE_ID


def device(devid=DEVICE_ID, status=STATUS.OPEN, low_battery=False, no_response=False):
    """Valve mock device."""
    return dict(
        actions=[
            {'label': 'Close', 'value': 'a=1&z=18&sw=off;'},
            {'label': 'Open', 'value': 'a=1&z=18&sw=on;'},
        ],
        area='1',
        bypass='0',
        control_url=CONTROL_URL,
        deep_link=None,
        default_group_id='1',
        faults={
            'low_battery': int(low_battery),
            'tempered': 0,
            'supervision': 0,
            'out_of_order': 0,
            'no_response': int(no_response),
        },
        group_id='xxxxx',
        group_name='Water leak',
        has_subscription=None,
        icon='assets/icons/water-value-shutoff.svg',
        id=devid,
        is_window='',
        name='Water shut-off valve',
        onboard='0',
        origin='abode',
        s2_dsk='',
        s2_grnt_keys='',
        s2_keys_valid='',
        s2_propty='',
        schar_24hr='0',
        sort_id='6',
        sort_order='',
        sresp_24hr='0',
        sresp_entry_0='0',
        sresp_entry_1='0',
        sresp_entry_2='0',
        sresp_entry_3='0',
        sresp_entry_4='0',
        sresp_exit_0='0',
        sresp_exit_1='0',
        sresp_exit_2='0',
        sresp_exit_3='0',
        sresp_exit_4='0',
        sresp_mode_0='0',
        sresp_mode_1='0',
        sresp_mode_2='0',
        sresp_mode_3='0',
        sresp_mode_4='0',
        status=status,
        status_color='#5cb85c',
        status_display='Open',
        status_ex='',
        status_icons=[],
        statuses={'switch': '1'},
        type='Shutoff Valve',
        type_tag='device_type.valve',
        uuid='xxxxxxxxxxxxxxxxxxxxxxxxx',
        version='021f00030002',
        zone='18',
        zwave_secure_protocol='',
    )



"""Mock Abode Glass Device."""

import jaraco.abode.devices.status as STATUS

DEVICE_ID = 'RF:00000001'


def device(
    devid=DEVICE_ID,
    status=STATUS.ONLINE,
    low_battery=False,
    no_response=False,
    out_of_order=False,
    tampered=False,
    uuid='91568b0d4c9d58c10d75fdeea887d4f4',
):
    """Glass break sensor mock device."""
    return dict(
        id=devid,
        type_tag='device_type.glass',
        type='GLASS',
        name='Glass Break Sensor',
        area='1',
        zone='6',
        sort_order=None,
        is_window='',
        bypass='0',
        schar_24hr='0',
        sresp_mode_0='0',
        sresp_entry_0='0',
        sresp_exit_0='0',
        sresp_mode_1='5',
        sresp_entry_1='5',
        sresp_exit_1='0',
        sresp_mode_2='5',
        sresp_entry_2='5',
        sresp_exit_2='0',
        sresp_mode_3='5',
        sresp_entry_3='5',
        sresp_exit_3='0',
        version='',
        origin='abode',
        control_url='',
        deep_link=None,
        status_color='#5cb85c',
        faults={
            'low_battery': int(low_battery),
            'tempered': int(tampered),
            'supervision': 0,
            'out_of_order': int(out_of_order),
            'no_response': int(no_response),
        },
        status=status,
        statuses={'hvac_mode': None},
        status_ex='',
        actions=[],
        status_icons=[],
        icon='assets/icons/unknown.svg',
        uuid=uuid,
    )
